'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.lib.exceptions import *
from arjuna.lib.types import constants
from arjuna.lib.utils import thread_utils
from arjuna.unitee.exceptions import *
from arjuna.unitee.enums import *
from arjuna.unitee.types.containers import CIStringDict
import datetime

class Issue:
    def __init__(self, *, iid, itype, ename, emsg, etrace):
        self.issue_timestamp = datetime.datetime.now().timestamp()
        self.iid = iid
        self.itype = itype
        self.ename = ename
        self.emsg = emsg
        self.etrace = etrace

class __BaseTestPoint:
    def __init__(self):
        self.__otype = None
        self.__tstamp = datetime.datetime.now().timestamp()
        self.__counter = "-"
        self.__exec_context = ExecutionContext.Test
        self.__step_type = StepType.ExecStep
        self.__purpose = constants.NOTSET_STRING
        self.__expectation = constants.NOTSET_STRING
        self.__observation = constants.NOTSET_STRING
        self.__assert_message = constants.NOTSET_STRING
        self.__has_failed = False
        self.__has_erred = False
        self.__iid = "-"
        self.__rtype = ResultTypeEnum.PASS
        self.__issue = None
        self.__step_props = "-"
        self.__source = "-"

        from arjuna.tpi import Arjuna
        self.unitee = Arjuna.get_unitee_instance()

    @property
    def otype(self):
        return self.__otype

    @otype.setter
    def otype(self, o_type):
        self.__otype = o_type

    @property
    def step_props(self):
        return self.__step_props

    @step_props.setter
    def step_props(self, props):
        self.__step_props = props and CIStringDict(props) or "-"

    @property
    def source(self):
        return self.__source

    @source.setter
    def source(self, source):
        self.__source = source and source or "-"

    @property
    def timestamp(self):
        return self.__tstamp

    @property
    def exec_context(self):
        return self.__exec_context

    @exec_context.setter
    def exec_context(self, exec_context):
        self.__exec_context = exec_context


    @property
    def step_type(self):
        return self.__step_type

    @step_type.setter
    def step_type(self, step_type):
        self.__step_type = step_type

    @property
    def issue(self):
        return self.__issue

    @issue.setter
    def issue(self, issue):
        self.__issue = issue

    def has_issue(self):
        return self.__issue is not None

    @property
    def iid(self):
        return self.__iid

    @iid.setter
    def iid(self, iid):
        self.__iid = iid

    @property
    def rtype(self):
        return self.__rtype

    @rtype.setter
    def rtype(self, rtype):
        self.__rtype = rtype

    @property
    def step_num(self):
        return self.__counter

    @step_num.setter
    def step_num(self, num):
        self.__counter = num

    @property
    def purpose(self):
        return self.__purpose

    @purpose.setter
    def purpose(self, purpose):
        if purpose is not None:
            self.__purpose = purpose

    def __trim_str(self, val):
        if val is not None and len(val) > 25000:
            return messages.Common.STR_LIMIT
        else:
            return val

    @property
    def expectation(self):
        return self.__expectation

    @expectation.setter
    def expectation(self, expectation):
        self.__expectation = self.__trim_str(expectation)

    @property
    def observation(self):
        return self.__observation

    @observation.setter
    def observation(self, observation):
        self.__observation = self.__trim_str(observation)

    @property
    def assert_message(self):
        return self.__assert_message

    @assert_message.setter
    def assert_message(self, message):
        self.__assert_message = self.__trim_str(message)

    def passed(self):
        return ((not self.failed() and not self.erred()))

    def failed(self):
        return self.__has_failed

    def set_failure(self):
        self.__has_failed = True
        self.rtype = ResultTypeEnum.FAIL

    def erred(self):
        return self.__has_erred

    def set_error(self):
        self.__has_erred = True
        self.rtype = ResultTypeEnum.ERROR

    def add_to_state(self):
        if self.failed():
            self.unitee.state_mgr.get_current_thread_state().add_fstep(self)
        elif self.erred():
            self.unitee.state_mgr.get_current_thread_state().add_estep(self)
        else:
            self.unitee.state_mgr.get_current_thread_state().add_pstep(self)


    def evaluate(self):
        if self.failed() or self.erred():
            raise StepResultEvent(self)

class Step(__BaseTestPoint):
    def __init__(self, ptname=None):
        super().__init__()
        self.ptname = ptname and ptname or thread_utils.get_current_thread_name()



import random

from arjuna.setuext.guiauto.impl.element.base_element import BaseElement
from arjuna.setuext.guiauto.impl.element.guielement import GuiElement

class GuiMultiElement(BaseElement):
    
    def __init__(self, automator, emd, parent=None):
        super().__init__(automator, emd, parent)
        self.instance_count = 0
        self.__instances = None

    def find(self):
        self.parent_container.find_multielement(self)

    #Override
    def find_if_not_found(self):
        if not self.is_found():
            self.find()

    def set_instance_count(self, count):
        self.instance_count = count
        self.__instances = [_GuiPartialElement(self.get_automator(), self, i) for i in range(self.instance_count)]

    def get_instance_count(self):
        self.find_if_not_found()
        return self.instance_count

    def get_random_index(self):
        self.find_if_not_found()
        return random.randint(0, self.instance_count-1)

    def get_instance_at_index(self, index):
        self.find_if_not_found()
        return self.__instances[index]

    def get_instance_at_ordinal(self, ordinal):
        self.find_if_not_found()
        return self.__instances[ordinal-1]

    def get_instance_by_visible_text(self, text):
        texts = self.__get_all_texts()
        first_index = self.__find_first_text_index(texts, text)
        return self.get_instance_at_index(first_index)

    def get_instance_by_value(self, value):
        values = self.__get_all_values()
        first_index = self.__find_first_value_index(values, value)
        return self.get_instance_at_index(first_index)

    def wait_until_visible(self):
        self.find_if_not_found()
        # Should the logic be keptin Setu itself for this????

    def __return_attr_values(self, response):
        if "data" in response and "attrValues" in response["data"]:
            return response["data"]["attrValues"]
        else:
            return None

    def get_tag_names(self):
        self.find_if_not_found()
        return [instance.get_tag_name() for instance in self.__instances]

    def get_text_contents(self):
        self.find_if_not_found()
        return [instance.get_text_content() for instance in self.__instances]

    def get_values(self):
        self.find_if_not_found()
        return [instance.get_attr_value("value") for instance in self.__instances]

    def get_attr_values(self, attr):
        self.find_if_not_found()
        return [instance.get_attr_value(attr=attr) for instance in self.__instances]

    def are_selected(self):
        self.find_if_not_found()
        return [instance.is_selected() for instance in self.__instances]

    # getting index attribute when it does not exist retursn value attribute.
    # So, not going the Selenium way. Setu would treat index as computer counting.
    def has_index_selected(self, index):
        self.find_if_not_found()
        return self.get_instance_at_index(index).is_selected()

    # Ordinal is human counting
    def has_ordinal_selected(self, ordinal):
        return self.has_index_selected(ordinal-1)

    def __find_first_match_index(self, in_sequence, to_match):
        try:
            return in_sequence.index(to_match)
        except:
            return -1

    def __get_all_texts(self):
        self.find_if_not_found()
        texts = self.get_text_contents()
        return texts

    def __find_first_text_index(self, texts, text):
        first_index = texts.index(text)
        if first_index == -1:
            raise Exception("No option with {} visible text present in drop down.".format(text))
        return first_index

    def __get_all_values(self):
        self.find_if_not_found()
        values = self.get_values()
        return values

    def __find_first_value_index(self, values, value):
        first_index = values.index(value)
        if first_index == -1:
            raise Exception("No option with {} value present in drop down.".format(value))
        return first_index

    def get_first_selected_instance(self):
        self.find_if_not_found()
        booleans = self.are_selected()
        first_index = None
        try:
            first_index = booleans.index(True)
        except:
            raise Exception("No option in drop down is currenlty selected.")
        else:
            return self.get_instance_at_index(first_index)

class _GuiPartialElement(GuiElement):

    def __init__(self, automator, multi_element: GuiMultiElement, instance_number: int):
        super().__init__(automator, multi_element.get_locator_meta_data())
        self.__multi_element = multi_element
        self.__instance_number = instance_number
        dispatcher = self.__multi_element.dispatcher_creator.create_gui_element_dispatcher(
            self.__multi_element.get_automator().dispatcher, self.__multi_element.setu_id
        )
        dispatcher.set_partial(self.__instance_number)
        self._set_dispatcher(dispatcher)

    def create_dispatcher(self):
        pass

    #Override
    def find_if_not_found(self):
        # Unlike regular item no attempt should be made to indepently identify a part element.
        pass

    def _is_partial_element(self):
        return True

    def _get_instance_number(self):
        return self.__instance_number


from enum import Enum, auto

class ArjunaOption(Enum):
    ARJUNA_ROOT_DIR = auto()
    ARJUNA_EXTERNAL_TOOLS_DIR = auto()
    ARJUNA_EXTERNAL_IMPORTS_DIR = auto()
    PYTHON_LOG_NAME = auto()
    LOG_NAME = auto()

    LOG_DIR = auto()

    LOG_CONSOLE_LEVEL = auto()
    LOG_FILE_LEVEL = auto()

    PROJECT_NAME = auto()
    PROJECT_ROOT_DIR = auto()
    PROJECT_CONF_FILE = auto()

    DATA_DIR = auto()
    DATA_SOURCES_DIR = auto()
    DATA_REFERENCES_DIR = auto()
    SCREENSHOTS_DIR = auto()
    CONFIG_DIR = auto()

    SETU_PROJECT_DIRS_FILES = auto()
    REPORT_DIR = auto()
    ARCHIVES_DIR = auto()

    AUT_URL = auto()

    TESTRUN_ENVIRONMENT = auto()
    TESTRUN_HOST_OS = auto()

    SETU_GUIAUTO_ACTOR_MODE = auto()
    SETU_GUIAUTO_ACTOR_URL = auto()

    AUTOMATOR_NAME = auto()

    BROWSER_NAME = auto()
    BROWSER_VERSION = auto()
    BROWSER_MAXIMIZE = auto()
    BROWSER_DIM_HEIGHT = auto()
    BROWSER_DIM_WIDTH = auto()
    BROWSER_BIN_PATH = auto()
    BROWSER_PROXY_ON = auto()

    GUIAUTO_INPUT_DIR = auto()
    GUIAUTO_NAMESPACE_DIR = auto()
    GUIAUTO_CONTEXT = auto()
    SCROLL_PIXELS = auto()
    SWIPE_TOP = auto()
    SWIPE_BOTTOM = auto()
    SWIPE_MAX_WAIT = auto()
    GUIAUTO_MAX_WAIT = auto()
    GUIAUTO_SLOMO_ON = auto()
    GUIAUTO_SLOMO_INTERVAL = auto()

    MOBILE_OS_NAME = auto()
    MOBILE_OS_VERSION = auto()
    MOBILE_DEVICE_NAME = auto()
    MOBILE_DEVICE_UDID = auto()
    MOBILE_APP_FILE_PATH = auto()

    SELENIUM_DRIVER_PROP = auto()
    SELENIUM_DRIVERS_DIR = auto()
    SELENIUM_DRIVER_PATH = auto()

    APPIUM_HUB_URL = auto()
    APPIUM_AUTO_LAUNCH = auto()

    IMAGE_COMPARISON_MIN_SCORE = auto()

    UNITEE_PROJECT_DIRS_FILES = auto()
    UNITEE_PROJECT_SESSIONS_DIR = auto()
    UNITEE_PROJECT_GROUPS_DIR = auto()
    UNITEE_PROJECT_TESTS_DIR = auto()
    UNITEE_PROJECT_TEST_MODULE_IMPORT_PREFIX = auto()
    UNITEE_PROJECT_FIXTURES_IMPORT_PREFIX = auto()
    UNITEE_PROJECT_CORE_DIR = auto()
    UNITEE_PROJECT_CORE_DB_CENTRAL_DIR = auto()
    UNITEE_PROJECT_CORE_DB_CENTRAL_DBFILE = auto()
    UNITEE_PROJECT_CORE_DB_RUN_DBFILE = auto()
    UNITEE_PROJECT_CORE_DB_ALLRUN_DIR = auto()
    UNITEE_PROJECT_CORE_DB_TEMPLATE_DIR = auto()
    UNITEE_PROJECT_CORE_DB_TEMPLATE_CENTRAL_DBFILE = auto()
    UNITEE_PROJECT_CORE_DB_TEMPLATE_RUN_DBFILE = auto()
    UNITEE_PROJECT_REPORTER_MODE = auto()
    UNITEE_PROJECT_ACTIVE_REPORTERS = auto()
    UNITEE_PROJECT_DEFERRED_REPORTERS = auto()
    UNITEE_PROJECT_FAILFAST = auto()
    UNITEE_PROJECT_REPORT_NAME_FORMAT = auto()
    UNITEE_PROJECT_REPORT_HEADERS_TMETA = auto()
    UNITEE_PROJECT_REPORT_HEADERS_IGMETA = auto()
    UNITEE_PROJECT_REPORT_HEADERS_PROPS = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_TEST = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_STEP = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_ISSUE = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_IGNORED = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_FIXTURE = auto()
    UNITEE_PROJECT_REPORT_HEADERS_REPORTABLE_EVENT = auto()
    UNITEE_PROJECT_RUNID = auto()
    UNITEE_PROJECT_IRUNID = auto()
    UNITEE_PROJECT_SESSION_NAME = auto()
    UNITEE_PROJECT_CORE = auto()
    UNITEE_PROJECT_SCREENSHOTS_RUN_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JDB_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_TESTS_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_IGNOREDTESTS_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_ISSUES_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_EVENTS_DIR = auto()
    UNITEE_PROJECT_RUN_REPORT_JSON_FIXTURES_DIR = auto()

class SetuActorMode(Enum):
    LOCAL = auto()
    REMOTE = auto()

class GuiAutomatorName(Enum):
    SELENIUM = auto()

class GuiAutomationContext(Enum):
	WEB = auto()
	NATIVE = auto()
	SCREEN = auto()
	ANDROID_WEB = auto()
	IOS_WEB = auto()
	ANDROID_NATIVE = auto()
	IOS_NATIVE = auto()

class BrowserName(Enum):
    FIREFOX = auto()
    SAFARI = auto()
    CHROME = auto()

class MobileOsName(Enum):
    ANDROID = auto()
    IOS = auto()

class DesktopOS(Enum):
    WINDOWS = auto()
    MAC = auto()
    LINUX = auto()


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import xml.etree.ElementTree as ETree

from arjuna.unitee.markup import mrules
from arjuna.unitee.enums import *

from .common.enums import *
from .common.rule import *
from .common.ref import *
from .bugs import *
from .tags import *
from .builtin_prop import *
from .user_prop import *
from .evars import *


class RuleBuilder:

    def __init__(self):
        self.__xml = None
        self.__totype = None
        self.__is_include_type = RuleNature.INCLUDE
        self.__target = None
        self.__robject = None
        self.__condition = None
        self.__expression = None

    def xml(self, xml):
        self.__xml = xml

    def test_object_type(self, totype):
        self.__totype = TestObjectTypeEnum[totype]
        return self

    def rule_type(self, rtype):
        self.__is_include_type = RuleNature[rtype.upper().strip()] == RuleNature.INCLUDE
        return self

    def target(self, container):
        self.__target = container

    def rule_object(self, target_name):
        self.__robject = target_name
        if type(self.__robject) is str:
            self.__robject = self.__robject.lower()
        return self

    def condition(self, condition):
        try:
            if type(condition) is str:
                if condition.lower() in SYMBOLS_MAP:
                    condition = SYMBOLS_MAP[condition.lower()]
                self.__condition = RuleConditionType[condition.upper()]
            else:
                self.__condition = condition
        except:
            raise Exception("Invalid condition supplied in rule: " + condition)
        return self

    def expression(self, expression):
        self.__expression = expression
        return self

    def build(self):
        if not self.__totype:
            raise Exception("Test object not defined for rule.")
        elif not self.__target:
            raise Exception("Target container not defined for rule.")
        elif not self.__robject:
            raise Exception("Target object not defined for rule.")
        elif not self.__condition:
            raise Exception("Condition not defined for rule.")
        elif not self.__expression:
            raise Exception("Expression not defined for rule.")

        if type(self.__robject) is str:
            if mrules.is_builtin_prop(self.__robject.upper()):
                # For built-in properties, checking for valid conditions happens at time of loading
                # For user defined properties, it is done as per the type of user defined property for a given test object
                # So, the latter can happen not at rule definition level, but based on test object
                # provided at run time.
                converter = mrules.get_value_type(self.__robject)
                allowed_condition_set = ALLOWED_CONDITIONS[converter]
                if self.__condition not in allowed_condition_set:
                    raise Exception("Rules for Built-in property >>{}<< of type {} can only use {} relation(s).".format(
                        self.__robject,
                        converter,
                        ALLOWED_SYMBOLS[converter]
                    ))

            # The condition has been named as contains for eace of use for string conditions in XML
            # Internally, contains for iterables means element content
            # For string it means partial match.
            if self.__condition == RuleConditionType.CONTAINS:
                self.__condition = RuleConditionType.PARTIALLY_MATCHES

        cls = None
        if self.__target == RuleTargetType.TAGS:
            cls = TagsDefinedRule
        elif self.__target == RuleTargetType.BUGS:
            cls = BugsDefinedRule
        elif self.__target == RuleTargetType.PROPS:
            if mrules.is_builtin_prop(self.__robject.upper()):
                if self.__expression in {'defined', 'present'}:
                    cls = BuiltInPropsDefinedRule
                else:
                    cls = BuiltInPropValueRule
            else:
                if self.__expression in {'defined', 'present'}:
                    cls = UserPropsDefinedRule
                else:
                    cls = UserPropValueRule

        elif self.__target == RuleTargetType.EVARS:
            if self.__expression in {'defined', 'present'}:
                cls = EvarsDefinedRule
            else:
                cls = EvarValueRule

        try:
            return cls(
                self.__totype,
                self.__is_include_type,
                self.__robject,
                self.__condition,
                self.__expression
            )
        except Exception as e:
            raise Exception("Problem in constructing rule object for {}. Message: {}".format(
                ETree.tostring(self.__xml).strip(),
                e
            ))


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.unitee.enums import *

class Kallable:

	def __init__(self, obj):
		self.obj = obj
		self.process()
		self.args = []

	def process(self):
		*pkg_parts, self.module = self.obj.__module__.split(".")
		self.pkg = ".".join(pkg_parts)
		self.name = self.obj.__name__
		self.qname = ".".join([self.pkg, self.module, self.name])

	def __call__(self):
		self.obj(*self.args)

class TestModule(Kallable):
	def process(self):
		*pkg_parts, self.module = self.obj.__name__.split(".")
		self.pkg = ".".join(pkg_parts)
		self.name = self.module
		self.qname = self.obj.__name__

class _TestFunction(Kallable):
	pass

class _DataSourceClass(Kallable):
	pass

class _DataSourceFunction(Kallable):
	pass

class _FixtureFunction(Kallable):
	def __init__(self, func, dec_name):
		super().__init__(func)
		self.type = FixtureTypeEnum[dec_name.upper()]
		self.func = func

class _DepModule(Kallable):
	pass

class _DepFunction(Kallable):
	pass

def __str(obj):
	try:
		return ".".join([obj.__module__, obj.__qualname__])
	except:
		return str(obj)

def create_test_module(obj):
	return TestModule(obj)

def create_dsource_class(obj):
	return _DataSourceClass(obj)

def create_dsource_func(obj):
	return _DataSourceFunction(obj)

def create_fixture(dec_name, obj):
	return _FixtureFunction(obj, dec_name)

def create_test_func(obj):
	return _TestFunction(obj)






from enum import Enum, auto

class ConfigActionType(Enum):
	REGISTER_NEW_CONFIG = auto()
	GET_ARJUNA_OPTION_VALUE = auto()
	GET_USER_OPTION_VALUE = auto()


class DataSourceActionType(Enum):
	CREATE_FILE_DATA_SOURCE = auto()
	GET_NEXT_RECORD = auto()
	GET_ALL_RECORDS = auto()
	RESET = auto()

class GuiActionType(Enum):
	CREATE_GUI = auto()
	CREATE_CHILD_GUI = auto()

class GuiAutoActionType(Enum):
    LAUNCH_AUTOMATOR = auto()
    QUIT_AUTOMATOR = auto()

    DEFINE_ELEMENT = auto()
    DEFINE_MULTIELEMENT = auto()
    DEFINE_DROPDOWN = auto()
    DEFINE_RADIOGROUP = auto()
    DEFINE_ALERT = auto()

    BROWSER_GO_TO_URL = auto()
    BROWSER_GO_BACK = auto()
    BROWSER_GO_FORWARD = auto()
    BROWSER_REFRESH = auto()
    BROWSER_EXECUTE_JAVASCRIPT = auto()

    CREATE_ELEMENT = auto()
    CREATE_MULTIELEMENT = auto()
    CREATE_DROPDOWN = auto()
    CREATE_RADIOGROUP = auto()
    CREATE_ALERT = auto()

    DEFINE_MAIN_WINDOW = auto()
    DEFINE_DOM_ROOT = auto()
    SET_SLOMO = auto()

    ELEMENT_ENTER_TEXT = auto()
    ELEMENT_SET_TEXT = auto()
    ELEMENT_CLEAR_TEXT = auto()

    ELEMENT_CLICK = auto() 

    ELEMENT_IDENTIFY = auto()
    ELEMENT_WAIT_UNTIL_PRESENT = auto()
    ELEMENT_WAIT_UNTIL_VISIBLE = auto()
    ELEMENT_WAIT_UNTIL_CLICKABLE = auto() 

    ELEMENT_CHECK = auto()
    ELEMENT_UNCHECK = auto()

    ELEMENT_GET_SOURCE = auto()

    MULTIELEMENT_GET_INSTANCE_COUNT = auto()
    MULTIELEMENT_GET_RANDOM_INDEX = auto()

    DROPDOWN_HAS_VALUE_SELECTED = auto()
    DROPDOWN_HAS_INDEX_SELECTED = auto()
    DROPDOWN_SELECT_BY_VALUE = auto()
    DROPDOWN_SELECT_BY_INDEX = auto() 
    DROPDOWN_GET_FIRST_SELECTED_OPTION_VALUE = auto()
    DROPDOWN_HAS_VISIBLE_TEXT_SELECTED = auto()
    DROPDOWN_GET_FIRST_SELECTED_OPTION_TEXT = auto()
    DROPDOWN_SELECT_BY_VISIBLE_TEXT = auto()

    RADIOGROUP_HAS_VALUE_SELECTED = auto()
    RADIOGROUP_HAS_INDEX_SELECTED = auto()
    RADIOGROUP_SELECT_BY_VALUE = auto()
    RADIOGROUP_SELECT_BY_INDEX = auto() 
    RADIOGROUP_GET_FIRST_SELECTED_OPTION_VALUE = auto()

    DOMROOT_FOCUS = auto()
    DOMROOT_CREATE_FRAME = auto()

    ALERT_CONFIRM = auto()
    ALERT_DISMISS = auto()
    ALERT_GET_TEXT = auto()
    ALERT_SEND_TEXT = auto()

    FRAME_FOCUS = auto()
    FRAME_CREATE_FRAME = auto()
    FRAME_GET_PARENT = auto()

    WINDOW_FOCUS = auto()
    WINDOW_GET_TITLE = auto() 

    MAIN_WINDOW_MAXIMIZE = auto()
    MAIN_WINDOW_CREATE_CHILD_WINDOW = auto()
    MAIN_WINDOW_GET_LATEST_CHILD_WINDOW = auto()
    MAIN_WINDOW_CLOSE_ALL_CHILD_WINDOWS = auto()

    CHILD_WINDOW_CLOSE = auto()

class SetuActionType(Enum):
	HELLO = auto()
	EXIT = auto()

class TestSessionActionType(Enum):
	INIT = auto()
	END = auto()


from .handler import Handler
from arjuna.setuext.guiauto.impl.base.conditions import *

class GuiElementConditions(Handler):

    def __init__(self, element):
        super().__init__(element)

    def IsSelected(self):
        caller = DynamicCaller(self.element.is_selected)
        return BooleanCondition(caller, True)

    def IsVisible(self):
        caller = DynamicCaller(self.element.is_visible)
        return BooleanCondition(caller, True)   

    def IsClickable(self):
        caller = DynamicCaller(self.element.is_clickable)
        return BooleanCondition(caller, True)     


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc

from arjuna.unitee.enums import *
from arjuna.unitee.markup import mrules

from .enums import *
from .exceptions import *
from .ref import *

class __BaseRule(metaclass=abc.ABCMeta):
    def __init__(self, rule_type, totype, is_inclusion_rule, robject, condition, expression):
        self.rule_type = rule_type
        self.totype = totype
        self.is_include_type = is_inclusion_rule
        self.robject = robject
        self.condition = condition
        self.expression = expression

        self._act_on_incompatible_converter(self.expression, name=self.robject)

    def _act_on_incompatible_converter(self, provided_value, name=None):
        pass

    def _not_met_exc(self):
        # Run time rules checking did not succeed for the test object
        raise RuleNotMet()

    def _raise_exception_for_check_result(self, check_result):
        if check_result:
            if self.is_include_type:
                return
            else:
                self._not_met_exc()
        else:
            if self.is_include_type:
                self._not_met_exc()
            else:
                return

    @abc.abstractmethod
    def _get_container(self, test_object):
        pass

    @abc.abstractmethod
    def is_object_defined_in(self, container):
        pass

    @abc.abstractmethod
    def _act_on_not_defined(self, expression):
        pass

    def _act_on_none_value(self, expression):
        pass

    def _get_target_value(self, container, name):
        return None

    def _act_on_value(self):
        pass

    def _convert_provided_value(self, provided_value, name=None, target_object_value=None):
        return provided_value

    def __is_applicable(self, test_object):
        if test_object.type not in {TestObjectTypeEnum.Module, TestObjectTypeEnum.Function}:
            return False
        elif test_object.type != self.totype:
            return False

        return True

    def evaluate(self, test_object):
        if not self.__is_applicable(test_object):
            return

        container = self._get_container(test_object)

        defined = self.is_object_defined_in(container)
        if not defined:
            self._act_on_not_defined(self.expression)

        if self.rule_type == RuleType.SET:
            return

        target_object_value = None
        if defined:
            target_object_value = self._get_target_value(container, self.robject)

        if target_object_value is None:
            self._act_on_none_value(self.expression)
            return

        if self.rule_type == RuleType.DICT_KEY:
            return

        try:
            provided_value = self._convert_provided_value(self.expression, name=self.robject, target_object_value=target_object_value)
        except Exception as e:
            # This would take place for user defined props/evars at this stage
            # E.g. in test user_prop is of type int and in rule it is given as a string
            # TO DO: Add more info and log as a run time error
            self._raise_exception_for_check_result(False)
        else:
            succeeded = self.__call_checker(target_object_value, provided_value)
            self._raise_exception_for_check_result(succeeded)

    def __call_checker(self, actual, expected):
        if type(actual) is str:
            actual = actual and actual.lower() or None
            expected = expected and expected.lower() or None
        func = VALUE_CHECKERS[self.condition]
        return func(actual, expected)

class _DictRule(__BaseRule):

    def __init__(self, rule_type, totype, is_inclusion_rule, robject, condition, expression):
        super().__init__(rule_type, totype, is_inclusion_rule, robject, condition, expression)

    def __contains(self, container, value):
        try:
            container[value]
            return True
        except:
            return False

    def is_object_defined_in(self, container):
        if type(self.robject) is str:
            return self.__contains(container, self.robject.lower())
        else:
            for value in self.robject:
                if not self.__contains(container, value):
                    return False
            return True

    def _act_on_not_defined(self, expression):
        if expression.lower() in {"defined", "present"}:
            self._raise_exception_for_check_result(False)
        elif expression.lower() == "none":
            # For a None value check, should be considered a match
            self._raise_exception_for_check_result(True)
        else:
            # For non-None expression, should be considered a non-match
            self._raise_exception_for_check_result(False)

    def _get_target_value(self, container, name):
        return container[name]

    def _act_on_none_value(self, expression):
        # Decision process is same as Not-def value. However this one works for dictionaries as well.
        self._act_on_not_defined(expression)

class DictKeyPresenceRule(_DictRule):

    def __init__(self, totype, is_inclusion_rule, robject, condition, expression):
        super().__init__(RuleType.DICT_KEY, totype, is_inclusion_rule, robject, condition, expression)

class DictKeyValueRule(_DictRule):

    def __init__(self, totype, is_inclusion_rule, robject, condition, expression):
        super().__init__(RuleType.DICT_VALUE, totype, is_inclusion_rule, robject, condition, expression)

class SetEntryPresenceRule(__BaseRule):

    def __init__(self, totype, is_inclusion_rule, robject, condition, expression):
        super().__init__(RuleType.SET, totype, is_inclusion_rule, robject, condition, expression)

    def is_object_defined_in(self, container):
        if type(self.robject) is str:
            return self.robject.lower() in container
        else:
            for i in self.robject:
                if i not in container:
                    return False
            return True

    def _act_on_not_defined(self, expression):
        self._raise_exception_for_check_result(False)





import abc

from arjuna.setuext.guiauto.impl.base.element_container import ElementContainer

class BaseElement(ElementContainer, metaclass=abc.ABCMeta):
    def __init__(self, automator, emd, parent=None):
        super().__init__(automator.config)
        self.__automator = automator
        self.__parent = parent
        self.__emd = emd
        self.__found = False
        self.__located_by = None 
        self.dispatcher_creator = automator.dispatcher_creator

    @property
    def parent_container(self):
        return self.__parent and self.__parent or self.__automator

    def _create_element_flat_or_nested(self, locator_meta_data):
        from arjuna.setuext.guiauto.impl.element.guielement import GuiElement
        return GuiElement(self.__automator, locator_meta_data, parent=self) 

    def _create_multielement_flat_or_nested(self, locator_meta_data):
        from arjuna.setuext.guiauto.impl.element.multielement import GuiMultiElement
        return GuiMultiElement(self.__automator, locator_meta_data, parent=self) 

    def create_dispatcher(self):
        self._set_dispatcher(self.dispatcher_creator.create_gui_element_dispatcher(self.__automator.dispatcher, self.setu_id))

    @abc.abstractmethod
    def find_if_not_found(self):
        pass

    def get_locator_meta_data(self):
        return self.__emd

    def is_found(self):
        return self.__found

    def set_found_with(self, locator_type, locator_value):
        self.__found = True
        self.__located_by = locator_type, locator_value

    def get_found_with(self, locator_type, locator_value):
        return self.__located_by

    def get_automator(self):
        return self.__automator



class DriverElementCommands:

    @classmethod
    def send_text(cls, element, text):
        element.send_keys(text)

    @classmethod
    def clear_text(cls, element):
        element.clear()

    @classmethod
    def submit(cls, element):
        element.submit()

    @classmethod
    def click(cls, element):
        element.click()

    @classmethod
    def get_text_content(cls, element):
        return element.text

    @classmethod
    def get_attr_value(cls, element, attr):
        return element.get_attribute(attr)

    @classmethod
    def get_tag_name(cls, element):
        return element.tag_name

    @classmethod
    def is_visible(cls, element):
        return element.is_displayed()

    @classmethod
    def is_clickable(cls, element):
        return element.is_enabled()

    @classmethod
    def is_selected(cls, element):
        return element.is_selected()


from arjuna.setu.types import SetuManagedObject
from .multielement import GuiMultiElement

# UUID is for client reference. Agent does not know about this.
class GuiWebRadioGroup(SetuManagedObject):

    def __init__(self, element_container, locator_meta_data):
        super().__init__()
        self._radios = element_container.create_multielement(locator_meta_data)
        self._validate_radio_buttons()

    def _validate_radio_buttons(self):
        tags = self._radios.get_tag_names()
        if [t for t in tags if t.strip().lower() != 'input']:
            raise Exception("Not a valid radio group. Contains non-input elements.")
        types = self._radios.get_attr_values("type")
        if [t for t in types if t.strip().lower() != 'radio']:
            raise Exception("Not a valid radio group. Contains non-radio elements.")
        names = self._radios.get_attr_values("name")
        refer_name = names[0]
        if [n for n in names if n != refer_name]:
            raise Exception("Not a valid radio group. Contains radio elements belonging to different radio groups.")

    def has_index_selected(self, index):
        return self._radios.get_instance_at_index(index).is_selected()

    def has_value_selected(self, value):
        return self._radios.get_instance_by_value(value).is_selected()

    def get_first_selected_option_value(self):
        instance = self._radios.get_first_selected_instance()
        return instance.get_attr_value("value")

    def select_by_index(self, index):
        self._radios.get_instance_at_index(index).select()

    def select_by_ordinal(self, ordinal):
        return self.select_by_index(ordinal-1)

    def select_by_value(self, value):
        return self._radios.get_instance_by_value(value).select()


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc


class TestNode(metaclass=abc.ABCMeta):

    def __init__(self, name):
        self.name = name
        self.edges = []
        self.scheduled = None

    def is_dummy(self):
        return False

    def add_edge(self, node):
        self.edges.append(node)

    def get_edges(self):
        return self.edges

    def enumerate(self):
        for edge in self.edges:
            if edge is None:
                ArjunaCore.console.display("None")
            else:
                ArjunaCore.console.display(edge.name)

class NonTestNode(TestNode):
    def __init__(self):
        super().__init__("Dummy")

    def is_dummy(self):
        return True


class DependencyTree:
    def __init__(self, names, nodes_map):
        self.name_queue = names
        self.name_set = {i for i in nodes_map}
        self.all_nodes_map = nodes_map

    def dep_exists(self, name):
        return name in self.name_set

    def get_node_count(self):
        return len(self.name_queue)

    def is_empty(self):
        return not self.name_queue

    def __iter__(self):
        return iter(self.name_queue)

    def remove_dep(self, name):
        self.name_set.remove(name)

    def remove_all_from_queue(self, names):
        self.name_queue = [i for i in self.name_queue if i not in set(names)]

    def get_node(self, name):
        return self.all_nodes_map[name]

class DepTreeBuilder:
    def __init__(self):
        self.base_node = NonTestNode()
        self.all_nodes_map = {}
        self.dependency_map = {}

        self.all_filtered_methods = set()
        self.resolved = set()
        self.unresolved = set()

    def create_node(self, name, dependency):
        node = TestNode(name)
        self.all_nodes_map[name] = node
        if dependency is not None:
            self.dependency_map[name] = [du.name for du in dependency if not du.should_ignore()]
        else:
            self.dependency_map[name] = None

    def process_dependencies(self):
        for tname in self.all_nodes_map:
            tnode = self.all_nodes_map[tname]
            tdeps = self.dependency_map[tname]
            if tdeps:
                for tdep in tdeps:
                    tnode.add_edge(self.all_nodes_map[tdep])
            self.base_node.add_edge(tnode)

    def validate(self):
        self.__resolve_deps(self.base_node)


    def __resolve_deps(self, node):
        self.unresolved.add(node)
        edges = node.get_edges()
        if edges:
            for edge in edges:
                if not edge: continue
                if edge not in self.resolved:
                    if edge in self.unresolved:
                        raise Exception("Circular reference: {} {}".format(node.name, edge.name))
                    else:
                        self.__resolve_deps(edge)

        self.resolved.add(node)
        self.unresolved.remove(node)

    def __build_dep_tree(self, scheduled_names):
        node_map = {}
        for name in scheduled_names:
            node_map[name] = TestNode(name)

        for name in scheduled_names:
            source = self.all_nodes_map[name]
            target = node_map[name]
            for edge in source.get_edges():
                if not edge: continue
                if edge.name in node_map:
                    target.add_edge(node_map[edge.name])

        return DependencyTree(scheduled_names, node_map)

    def slot_names(self, scheduled_names):
        tree = self.__build_dep_tree(scheduled_names)
        current_slot_num = 1
        slots = []
        name_slot_map = {}
        done = []

        def is_done(node):
            edges = node.get_edges()
            if edges:
                return True
            else:
                for edge in edges:
                    if not edge: continue
                    if tree.dep_exists(edge.name):
                        return False
                    else:
                        if name_slot_map[edge.name] == current_slot_num:
                            return False
                return True

        while not tree.is_empty():
            slot = []
            for name in tree:
                node = tree.get_node(name)
                dep_met = is_done(node)
                if dep_met:
                    slot.append(name)
                    name_slot_map[name] = current_slot_num
                    tree.remove_dep(name)
                    done.append(name)

            tree.remove_all_from_queue(done)
            current_slot_num += 1
            slots.append(slot)

        return slots








import re
import os

from enum import Enum, auto

from arjuna.tpi.enums import GuiAutomationContext
from arjuna.setuext.guiauto.impl.locator.emd import GuiElementMetaData, Locator

class FileFormat(Enum):
    GNS = auto()
    XLS = auto()
    XLSX = auto()

class GuiNamespaceLoaderFactory:

    # Returns GuiNamespaceLoader
    @classmethod
    def create_namespace_loader(cls, ns_file_path):
        _, file_extension = os.path.splitext(ns_file_path)
        ext = file_extension.upper()[1:]
        considered_path = ns_file_path
        try:
            file_format = FileFormat[ext]
        except:
            raise Exception("Unsupported format for namespace: {}".format(file_extension))
        else:
            full_file_path = ns_file_path
            if os.path.isdir(full_file_path):
                raise Exception("Namespace file path is a directory and not a file: {}".format(considered_path))
            elif not os.path.isfile(full_file_path):
                raise Exception("Namespace file path is invalid: {}".format(considered_path))

            if file_format == FileFormat.GNS:
                return NamespaceFileLoader(full_file_path)
            else:
                raise Exception("Unsupported format for namespace: {}".format(file_extension))


class GuiNamespace:

    def __init__(self, name):
        self.__name = name
        # dict <string, dict<GuiAutomationContext, GuiElementMetaData>>
        self.__ns = {}

    def add_element_meta_data(self, name, context, raw_locators):
        emd = GuiElementMetaData(raw_locators)
        name = name.lower()
        if not self.has(name):
            self.__ns[name] = {}
        self.__ns[name][context] = emd

    def has(self, name):
        return name.lower() in self.__ns

    def has_context(self, name, context):
        if self.has(name):
            return context in self.__ns[name.lower()]
        return False

    # Needs to be thread-safe
    # Returns emd for a context for a given gui name
    def get_meta_data(self, name, context):
        if not self.has(name):
            raise Exception("Gui namespace >{}< does not contain element with name: {}".format(self.__name, name))
        elif not self.has_context(name, context):
            raise Exception("Gui namespace >{}< does not contain element with name: {} for context {}".format(self.__name, name, context))
        
        return self.__ns[name.lower()][context]


class BaseGuiNamespaceLoader:

    def __init__(self, name):
        self.__name = name
        self.__namespace = GuiNamespace(name)

    @property
    def name(self):
        return self.__name

    @property
    def namespace(self):
        return self.__namespace

    # Needs to be thread safe
    def add_element_meta_data(self, name, context, locators):
        self.__namespace.add_element_meta_data(name, context, locators)

    def _raise_notafile_exception(self, file_path):
        raise Exception("{} is not a file.".format(file_path))

    def _raise_filenotfound_exception(self, file_path):
        raise Exception("{} is not a valid file path.".format(file_path))

    def _raise_relativepath_exception(self, file_path):
        raise Exception("Gui namespace loader does not accept relative file path. {} is not a full file path.".format(file_path))


class NamespaceFileLoader(BaseGuiNamespaceLoader):

    def __init__(self, ns_file_path):
        super().__init__(os.path.basename(ns_file_path))
        self.__ns_file = None
        self.__ns_path = None
        self.name_pattern = re.compile(r"\[\s*(.*?)\s*\]$")
        self.platform_pattern = re.compile(r"\s*\#\s*(.*?)\s*$")
        self.locator_pattern = re.compile(r"\s*(.*?)\s*=\s*(.*?)\s*$")
        self.header_found = False
        self.last_header = None
        self.last_auto_contexts = None # list of auto contexts
        #Map<String, Map<GuiAutomationContext,List<Locator>>>
        self.__ns = {}

        if not os.path.isabs(ns_file_path):
            super()._raise_relativepath_exception(ns_file_path)
        elif not os.path.exists(ns_file_path):
            super()._raise_filenotfound_exception(ns_file_path)
        elif not os.path.isfile(ns_file_path):
            super()._raise_notafile_exception(ns_file_path)

        self.__ns_path = ns_file_path
        self.__ns_file = open(self.__ns_path)

    def __match_header(self, input):
        match = self.name_pattern.match(input)
        if match:
            current_header = match.group(1)
            if not self.last_header:
                self.last_header = current_header
            elif self.last_header.lower() == current_header.lower():
                raise Exception("Found duplicate namespace definition for {} element.".format(self.last_header))
            else:
                if len(self.__ns[self.last_header]) == 0:
                    raise Exception("Found empty namespace definition for {} element.".format(self.last_header))
                else:
                    for context, data in self.__ns[self.last_header].items():
                        if len(data) == 0:
                            raise Exception("Found empty namespace definition for {} context for {} element.".format(context.name, self.last_header))
                self.last_header = current_header

            self.last_auto_contexts = None
            self.__ns[self.last_header] = {}
            return True
        else:
            return False

    def __match_contexts(self, input):
        match = self.platform_pattern.match(input)
        if match:        
            try:
                raw_contexts = [m.strip().upper() for m in match.group(1).split(",")]
                contexts = [GuiAutomationContext[n] for n in raw_contexts]
            except Exception as e:
                raise Exception("Invalid context name found in header: {}".format(e))
            else:
                for context in contexts:
                    if context in self.__ns[self.last_header]:
                        raise Exception("Found duplicate automation context {} in {} namespace definition.".format(context.name, self.last_header))
                    else:
                        self.__ns[self.last_header][context] = []
            
            self.last_auto_contexts = contexts
            return True
        else:
            return False

    def __match_locator(self, input):
        match = self.locator_pattern.match(input)
        if match:       
            locator = Locator(match.group(1), match.group(2))
            if (self.last_auto_contexts is None):
                raise Exception("Locators must be preceded with context information as #context1, context2 construct. Current line: " + input)   

            for context in self.last_auto_contexts:
                self.__ns[self.last_header][context].append(locator)
            return True
        else:
            return False

    def load(self):
        for line in self.__ns_file.readlines():
            line = line.strip()
            if not line: 
                continue
            if self.__match_header(line):
                self.header_found = True
                continue
            else:
                if not self.header_found:
                    raise Exception("Namespace contents must be contained inside a [name] header.")
                elif self.__match_contexts(line):
                    continue
                elif self.__match_locator(line):
                    continue
                else:
                    raise Exception("Unexpected namespace file entry. Namspace content can either be plaforms or identification definition: " + line)
        
        self.__ns_file.close()

        for ename, context_data in self.__ns.items():
            for context, locators in context_data.items():
                self.add_element_meta_data(ename, context, locators)


from arjuna.setu.types import SetuManagedObject
from arjuna.setuext.guiauto.impl.element.guielement import GuiElement
from arjuna.tpi.enums import ArjunaOption

class WebAlert(SetuManagedObject):

    def __init__(self, automator):
        super().__init__()
        self.__automator = automator

    @property
    def automator(self):
        return self.__automator

    def confirm(self):
        self.automator.dispatcher.confirm_web_alert()
        self.automator.alert_handler.delete_alert()

    def dismiss(self):
        self.automator.dispatcher.dismiss_web_alert()
        self.automator.alert_handler.delete_alert()

    def send_text(self, text):
        self.automator.dispatcher.send_text_to_web_alert(text)

    def get_text(self):
        return self.automator.dispatcher.get_text_from_web_alert()



class DriverCommands:

    @classmethod
    def go_to_url(cls, driver, url):
        driver.get(url)

    @classmethod
    def refresh_browser(cls, driver):
        driver.refresh()

    @classmethod
    def go_back_in_browser(cls, driver):
        driver.back()

    @classmethod
    def go_forward_in_browser(cls, driver):
        driver.forward()

    @classmethod
    def quit(cls, driver):
        driver.quit()

    @classmethod
    def get_page_title(cls, driver):
        return driver.title

    @classmethod
    def get_url(cls, driver):
        return driver

    @classmethod
    def get_page_source(cls, driver):
        return driver.page_source

    @classmethod
    def is_web_alert_present(cls, driver):
        from selenium.webdriver.support.ui import WebDriverWait
        from selenium.webdriver.support import expected_conditions as EC
        try:
            WebDriverWait(driver, 1).until(EC.alert_is_present(),'Timed out.')
            return True
        except Exception as e:
            return False

    @classmethod
    def confirm_web_alert(cls, driver):
        driver.switch_to.alert.accept()

    @classmethod
    def dismiss_web_alert(cls, driver):
        driver.switch_to.alert.dismiss()

    @classmethod
    def send_text_to_web_alert(cls, driver, text):
        driver.switch_to.alert.send_keys(text)

    @classmethod
    def get_text_from_web_alert(cls, driver):
        return driver.switch_to.alert.text

    @classmethod
    def focus_on_frame(cls, driver, element):
        driver.switch_to.frame(element)

    @classmethod
    def focus_on_dom_root(cls, driver):
        return driver.switch_to.default_content()

    @classmethod
    def focus_on_parent_frame(cls, driver):
        driver.switch_to.parent_frame()

    @classmethod
    def execute_javascript(cls, driver, script):
        return driver.execute_script(script)

    @classmethod
    def take_screenshot(cls, driver):
        return driver.get_screenshot_as_base64()

    @classmethod
    def set_window_size(cls, driver, width, height):
        driver.set_window_size(width, height)

    @classmethod
    def maximize_window(cls, driver):
        driver.maximize_window()

    @classmethod
    def get_current_window_handle(cls, driver):
        return driver.current_window_handle

    @classmethod
    def focus_on_window(cls, driver, window_handle):
        driver.switch_to.window(window_handle)

    @classmethod
    def close_current_window(cls, driver):
        driver.close()

    @classmethod
    def get_window_title(cls, driver):
        return driver.title

    @classmethod
    def get_current_window_size(cls, driver):
        return driver.get_window_size()

    @classmethod
    def get_all_winodw_handles(cls, driver):
        return driver.window_handles


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import configparser


class IniFile2MapReader:
    def __init__(self, file_path):
        self.file_path = file_path
        self._f = configparser.ConfigParser()
        self._f.read(self.file_path)
        self.sections = self._f.sections()

    def __get_section_data(self, section):
        return self._f[section]

    def __iter__(self):
        return self

    def next(self):
        if self.sections:
            return self.__get_section_data(self.sections.pop(0))
        else:
            raise StopIteration()

    def read(self):
        section_data = {}
        for section in self.sections:
            section_data[section] = dict(self._f[section])
        return section_data

    def close(self):
        self._f.clear()
        del self._f



from .driverelement import SeleniumDriverElement
from arjuna.setuext.guiauto.dispatcher.driver.impl.driver_commands import DriverCommands
from arjuna.setuext.guiauto.dispatcher.driver.impl.element_finder import ElementFinder


class SeleniumDriver:

    def __init__(self, setu_id):
        self.__setu_id = setu_id
        self.__config = None
        self.__driver = None
        self.__driver_elements = {}
        self.__driver_melements = {}

    def create_gui_element_dispatcher(self, element_setu_id):
        return SeleniumDriverElement(self, element_setu_id)

    @property
    def setu_id(self):
        return self.__setu_id

    @property
    def driver(self):
        return self.__driver

    def add_driver_element(self, setu_id, element):
        self.__driver_elements[setu_id] = element

    def add_driver_melement(self, setu_id, melement):
        self.__driver_melements[setu_id] = melement

    def get_driver_element(self, setu_id):
        return self.__driver_elements[setu_id]

    def get_driver_melement(self, setu_id):
        return self.__driver_melements[setu_id]

    def __create_success_response(self):
        response = dict()
        response["result"] = "success"
        response["data"] = {}
        return response

    def launch(self, config):
        self.__config = config
        from .impl.browser_launcher import BrowserLauncher
        self.__driver = BrowserLauncher.launch(config) 

    def quit(self):
        DriverCommands.quit(self.__driver)

    def go_to_url(self, url):
        DriverCommands.go_to_url(self.__driver, url)

    def go_back_in_browser(self):
        DriverCommands.go_back_in_browser(self.__driver)

    def go_forward_in_browser(self):
        DriverCommands.go_forward_in_browser(self.__driver)

    def refersh_browser(self):
        DriverCommands.refersh_browser(self.__driver)

    def execute_javascript(self, script):
        DriverCommands.execute_javascript(self.__driver, script)

    def take_screenshot(self):
        DriverCommands.take_screenshot(self.__driver)

    def find_element(self, child_gui_element_setu_id, with_type, with_value):
        element = ElementFinder.find_element(self.__driver, with_type, with_value)
        self.__driver_elements[child_gui_element_setu_id] = element

    def find_multielement(self, child_gui_element_setu_id, with_type, with_value):
        melement = ElementFinder.find_elements(self.__driver, with_type, with_value)
        self.__driver_melements[child_gui_element_setu_id] = melement
        return melement.get_instance_count()

    def get_current_window_handle(self):
        return DriverCommands.get_current_window_handle(self.__driver)

    def get_current_window_title(self):
        return DriverCommands.get_window_title(self.__driver)

    def maximize_current_window(self):
        DriverCommands.maximize_window(self.__driver)

    def get_current_window_size(self):
        res = DriverCommands.get_current_window_size(self.__driver)
        return {"width" : res[0], "height" : res[1]}

    def get_all_window_handles(self):
        return DriverCommands.get_all_winodw_handles(self.__driver)

    def focus_on_window(self, handle):
        DriverCommands.focus_on_window(self.__driver, handle)

    def close_current_window(self):
        DriverCommands.close_current_window(self.__driver)

    def is_web_alert_present(self):
        return DriverCommands.is_web_alert_present(self.__driver)

    def confirm_web_alert(self):
        DriverCommands.confirm_web_alert(self.__driver)

    def dismiss_web_alert(self):
        DriverCommands.dismiss_web_alert(self.__driver)

    def get_text_from_web_alert(self):
        return DriverCommands.get_text_from_web_alert(self.__driver)

    def send_text_to_web_alert(self, text):
        DriverCommands.send_text_to_web_alert(self.__driver, text)

    def focus_on_frame(self, elem_setu_id, is_instance_action=False, instance_index=0):
        element = None
        if is_instance_action:
            element = self.__driver_melements[elem_setu_id].get_element_at_index(instance_index)
        else:
            element = self.__driver_elements[elem_setu_id]
        DriverCommands.focus_on_frame(self.__driver, element)

    def focus_on_parent_frame(self):
        DriverCommands.focus_on_parent_frame(self.__driver)

    def focus_on_dom_root(self):
        DriverCommands.focus_on_dom_root(self.__driver)



from enum import Enum, auto

class SetuGuiAutoActorStateActionType(Enum):
    INIT_STATE = auto()
    END_STATE = auto()

class SetuGuiAutoActorAutomatorActionType(Enum):

    LAUNCH = auto()
    QUIT = auto()

    GO_TO_URL = auto()

    GO_BACK = auto()
    GO_FORWARD = auto()
    REFRESH = auto()
    EXECUTE_JAVASCRIPT = auto()

    FIND_ELEMENT = auto()
    FIND_MULTIELEMENT = auto()

    GET_CURRENT_WINDOW_HANDLE = auto()
    GET_WINDOW_TITLE = auto()
    MAXIMIZE_WINDOW = auto()
    GET_CURRENT_WINDOW_SIZE = auto()
    SET_WINDOW_SIZE = auto()
    GET_ALL_WINDOW_HANDLES = auto()
    SWITCH_TO_WINDOW = auto()
    CLOSE_CURRENT_WINDOW = auto()

    IS_ALERT_PRESENT = auto()
    CONFIRM_ALERT = auto()
    DISMISS_ALERT = auto()
    GET_TEXT_FROM_ALERT = auto()
    SEND_TEXT_TO_ALERT = auto()

    GET_CURRENT_VIEW_CONTEXT = auto()
    GET_ALL_VIEW_CONTEXTS = auto()
    SWITCH_TO_VIEW_CONTEXT = auto()

    JUMP_TO_FRAME = auto()
    JUMP_TO_PARENT_FRAME = auto()
    JUMP_TO_DOM_ROOT = auto()

class SetuGuiAutoActorElementActionType(Enum):
	FIND_MULTIELEMENT = auto()
	FIND_ELEMENT = auto()
	CLICK = auto()
	CLEAR_TEXT = auto()
	SEND_TEXT = auto()
	IS_SELECTED = auto()
	IS_VISIBLE = auto()
	IS_CLICKABLE = auto()
	GET_TAG_NAME = auto()
	GET_ATTR_VALUE = auto()
	GET_TEXT_CONTENT = auto()

class SetuActorDriverConfigOption(Enum):
    AUTOMATOR_NAME = auto()
    GUIAUTO_CONTEXT = auto()

    # Browser (Common)
    BROWSER_NAME = auto()
    BROWSER_BIN_PATH = auto()
    BROWSER_PROXY_ON = auto()
    BROWSER_PROXY_HOST = auto()
    BROWSER_PROXY_PORT = auto()
    MOBILE_OS_NAME = auto()

    # Selenium
    SELENIUM_DRIVER_PROP = auto()
    SELENIUM_DRIVER_PATH = auto()

    # Appium
    APPIUM_HUB_URL = auto()
    APPIUM_AUTO_LAUNCH = auto()


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.tpi.enums import ArjunaOption
from arjuna.lib.adv.decorators import singleton
from arjuna.lib.reader.hocon import *
from arjuna.lib.utils import sys_utils
from arjuna.unitee.enums import *
from arjuna.unitee.types.root import Root
from arjuna.unitee.test.defs.gconf import GroupConfsLoader, GroupConf
from arjuna.unitee.state.defdb import *
from arjuna.lib.utils import file_utils, thread_utils
from arjuna.lib.config import InternalTestContext

@singleton
class Unitee:

    def __init__(self, setu_test_session, config):
        super().__init__()
        self.__state_mgr = None
        self.__reporter = None
        self.__test_loader = None
        self.__tdb = None
        self.__session = None
        self.__cli_picker_options = None
        self.__frozen = False
        self.__groups = None
        self.__setu_test_session = setu_test_session
        self.__config = config

        if self.__frozen:
            raise Exception("Unitee has already been loaded.")

        from arjuna.unitee.state.states import ThreadManager
        from arjuna.unitee.state.reporter import ActiveReporter
        self.__state_mgr = ThreadManager()
        self.__reporter = ActiveReporter()
        self.__reporter.set_up()

        from arjuna.unitee.state.loader import TestLoader
        self.__test_loader = TestLoader()

    @property
    def state_mgr(self):
        return self.__state_mgr

    @property
    def reporter(self):
        return self.__reporter

    @property
    def session(self):
        return self.__session

    @property
    def test_loader(self):
        return self.__test_loader

    @property
    def groups(self):
        return self.__groups

    @property
    def testdb(self):
        return self.__tdb

    def load_testdb(self):
        self.__tdb = DefDB()
        self.test_loader.load()
        self.__groups = GroupConfsLoader.load()

    def load_session_for_all(self):
        from arjuna.unitee.test.defs.session import MSessionAllTests
        sdef = MSessionAllTests()
        sdef.process()
        self.__session = sdef.pick()
        self.__session.context = self.__create_test_context()
        self.__session.load()

    def __create_test_context(self):
        context = InternalTestContext(self.__setu_test_session, "default", parent_config=self.__config)
        # Populate default config (central config post project conf processing)
        context.ConfigBuilder(code_mode=False).build()
        return context

    def load_session(self, session_name):
        from arjuna.tpi import Arjuna
        from arjuna.unitee.test.defs.session import UserDefinedSessionDef
        sdir = Arjuna.get_central_config().get_arjuna_option_value(ArjunaOption.UNITEE_PROJECT_SESSIONS_DIR).as_string()
        session_file_path = os.path.join(sdir, session_name + ".xml")
        if not file_utils.is_file(session_file_path):
            Arjuna.get_console().display_error("Not able to find session file {}.xml at {}".format(session_name, sdir))
            sys_utils.fexit()
        sdef = UserDefinedSessionDef(session_name, session_file_path)
        sdef.process()
        self.__session = sdef.pick()
        self.__session.context = self.__create_test_context()
        self.__session.load()

    def load_session_for_group(self, group_name):
        from arjuna.unitee.test.defs.session import MSessionSingleGroup
        sdef = MSessionSingleGroup(group_name)
        sdef.process()
        self.__session = sdef.pick()
        self.__session.context = self.__create_test_context()
        self.__session.load()

    def load_session_for_name_pickers(self, **picker_args):
        from xml.etree.ElementTree import Element, SubElement, Comment, tostring
        gname = '_mngroup'
        group_xml = Element('group')
        group_xml.attrib['name'] = gname

        pickers = Element('pickers')

        pattern_list = ['cm', 'im', 'cf', 'if']

        for picker in pattern_list:
            patterns = picker_args[picker]
            if patterns is not None:
                for pattern in patterns:
                    child = Element('picker')
                    child.attrib['type'] = picker
                    child.attrib['pattern'] = pattern
                    pickers.append(child)

        group_xml.append(pickers)

        gconf = GroupConf(gname, group_xml, "(Arjuna generated object)")
        self.__groups[gname] = gconf
        self.load_session_for_group(gname)

    def run(self):
        self.state_mgr.register_thread(thread_utils.get_current_thread_name())
        self.__session.run()

    def tear_down(self):
        self.__reporter.tear_down()


'''
Dummy Remote implementation for testability purpose.

Depending on dummy area, it could be a fully/partially working dummy actor.

From testability aspect, it should assert whether expected packet was got. TO DO!!!
'''


import re
import xml.etree.ElementTree as ETree

from arjuna.tpi.enums import ArjunaOption
from arjuna.unitee.types.containers import *
from arjuna.unitee.enums import *
from arjuna.lib.utils import etree_utils
from arjuna.lib.utils import sys_utils
from arjuna.unitee.utils import run_conf_utils

class NameMatcher:

    def __init__(self, consider_names, ignore_names):
        self.__consider_names = consider_names
        self.__ignore_names = ignore_names

        # Ignore overrides consider
        for name in self.__ignore_names:
            if name in self.__consider_names:
                self.__consider_names.remove(name)

    def matches(self, name):
        if self.__consider_names:
            consider = False
            for reg in self.__consider_names:
                if reg.match(name):
                    consider = True
            return consider

        if self.__ignore_names:
            consider = True
            for reg in self.__ignore_names:
                if reg.match(name):
                    consider = False
            return consider

        return True


class Picker:

    def __init__(self, fpath, gname, pickers_node):
        self.gname = gname
        self.fpath = fpath
        self.root = pickers_node

        # Although using a set would give better performance, using a list so that pickers are evaluated
        # in the order they are specified.
        self._rule_dict = CIStringDict(
            {
                'cm': [],
                'im': [],
                'cf': [],
                'if': []
            }
        )

        self.__module_matcher = None
        self.__function_matcher = None

        from arjuna.tpi import Arjuna
        self.console = Arjuna.get_console()
        self.tm_prefix = Arjuna.get_central_config().get_arjuna_option_value(ArjunaOption.UNITEE_PROJECT_TEST_MODULE_IMPORT_PREFIX).as_string()
        self.__process()

    def __process(self):
        def display_err_and_exit(msg):
            self.console.display_error((msg + " Fix groups template file: {}").format(self.fpath))
            sys_utils.fexit()

        node_dict = etree_utils.convert_to_cidict(self.root)
        pattern_set = {i.lower() for i in self._rule_dict.keys()}

        # Validate only pickers keys exist.
        if node_dict.keys() != {'picker'}:
            display_err_and_exit(">>pickers<< element can contain only one or more >>picker<< element. Check group: {}".format(self.gname))
        elif not node_dict:
            display_err_and_exit(">>pickers<< element must contain atleast one >>picker<< element.Check group: {}".format(self.gname))

        for pattern in list(self.root):
            run_conf_utils.validate_pattern_xml_child("groups", self.gname, self.fpath, pattern)
            pattern_dict = etree_utils.convert_attribs_to_cidict(pattern)
            pattern_name = pattern_dict['type'].lower()
            if pattern_name not in pattern_set:
                display_err_and_exit(
                    ">>pickers<< element type attribute can only contain one of {}.Check group: {}".format(
                        pattern_set,
                        self.gname))
            pattern_value = pattern_dict['pattern'].strip()
            if not pattern_value:
                display_err_and_exit(
                    ">>pickers<< element's >>pattern<< attribute can not be empty.Check group: {}".format(
                        pattern_set,
                        self.gname))

            if pattern_name in {'cm', 'im'}:
                self._rule_dict[pattern_name].append(re.compile(self.tm_prefix + pattern_value))
            else:
                self._rule_dict[pattern_name].append(re.compile(pattern_value))

        self.__module_matcher = NameMatcher(self._rule_dict['cm'], self._rule_dict['im'])
        self.__function_matcher = NameMatcher(self._rule_dict['cf'], self._rule_dict['if'])


    def include_module(self, mname):
        return self.__module_matcher.matches(mname)


    def include_func(self, fname):
        return self.__function_matcher.matches(fname)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import datetime

from collections import OrderedDict

from arjuna.lib.adv.types import CIStringDict
from arjuna.unitee.enums import *
from arjuna.lib.utils import thread_utils

class DictWrapper:

    def __init__(self, header_enum, sdict=None):
        vars(self)['_d'] = OrderedDict()
        for henum in header_enum:
            self.__update(henum.name.lower(), "-", False)
        if sdict:
            self.update_from_dict(sdict)

    def __getattr__(self, item):
        if item.startswith("_"): return vars(self)[item]
        else:
            self.__validate_attr(item)
            return self._d[item.lower()]

    def __setattr__(self, attr, value):
        if attr.startswith("_"):
            vars(self)[attr] = value
            return
        self.__update(attr, value)

    def headers(self):
        return tuple(self._d.keys())

    def values(self):
        return tuple(self._d.values())

    def get_dict(self):
        return self._d

    def items(self):
        return self._d.items()

    def __validate_attr(self, k):
        if k.lower() not in self._d:
            raise AttributeError("{} is not an allowed result attribute for {}".format(k, self.__class__.__name__))

    def __update(self, k, v, validate=True):
        if validate:
            self.__validate_attr(k)
        self._d[k.lower()] = v

    def update_from_dict(self, d):
        for i,j in d.items():
            self.__update(i,j)

class ReportEntry(DictWrapper):
    def __init__(self, entry_enum, obj):
        super().__init__(entry_enum)
        self.otype = obj.type.name
        self.session = obj.tvars.info.session.meta["name"]
        if obj.tvars.info.stage:
            self.stage = obj.tvars.info.stage.meta["name"]
        if obj.tvars.info.group:
            self.group = obj.tvars.info.group.meta["name"]
        if obj.tvars.info.module:
            self.pkg = obj.tvars.info.module.meta["pkg"]
            self.module = obj.tvars.info.module.meta["name"]
        if obj.tvars.info.function:
            self.function = obj.tvars.info.function.meta["name"]
        if obj.tvars.info.test_num:
            self.test = obj.tvars.info.test_num

class Result(ReportEntry):

    def __init__(self, obj):
        super().__init__(ResultEntryEnum, obj)

    def is_stepped(self):
        return False

class StepResult(DictWrapper):
    def __init__(self):
        super().__init__(StepEntryEnum)

class IssueEntry(DictWrapper):
    def __init__(self):
        super().__init__(IssueEntryEnum)

class Info(ReportEntry):

    def __init__(self, info_type, obj):
        super().__init__(InfoEntryEnum, obj)
        self.info_timestamp = datetime.datetime.now().timestamp()
        self.message = info_type.name.lower()

class FixtureInfo(Info):
    def __init__(self, fix, info_type, obj):
        super().__init__(info_type, obj)
        self.otype = TestObjectTypeEnum.FIXTURE
        self.fixture = fix.fqname
        self.fixtype = fix.ftype

class TestObjectResult(Result):

    def __init__(self, test_obj, *, rtype=None, rcode=None, iid=None):
        super().__init__(test_obj)
        self.result_timestamp = datetime.datetime.now().timestamp()
        self.rtype = rtype and rtype.name or "-"
        self.rcode = rcode and rcode.name or "-"
        self.iid = iid and iid or "-"
        self._issues = []
        self._steps = []
        if not test_obj.tvars.evars.is_empty():
            self.evars = str(test_obj.tvars.evars)
        if not test_obj.tvars.tags.is_empty():
            self.tags = str(test_obj.tvars.tags)
        if test_obj.type == TestObjectTypeEnum.Module:
            if not test_obj.tvars.info.module.props.is_empty():
                self.props = test_obj.tvars.info.module.props
        elif test_obj.type in {TestObjectTypeEnum.Function, TestObjectTypeEnum.Test}:
            if not test_obj.tvars.info.module.props.is_empty() and not test_obj.tvars.info.function.props.is_empty():
                parts = []
                parts.append("Module Props")
                parts.append(str(test_obj.tvars.info.module.props))
                parts.append("Function Props")
                parts.append(str(test_obj.tvars.info.function.props))
                self.props = "\n".join(parts)
        if not test_obj.tvars.runtime.is_empty():
            self.runtime = str(test_obj.tvars.runtime)
        self.thread_id = thread_utils.get_current_thread_name()

    @property
    def steps(self):
        return self._steps

    def is_stepped(self):
        return False

    @property
    def issues(self):
        return self._issues

    def has_issues(self):
        return len(self._issues) !=0

    def append_issue(self, issue):
        issue_entry = IssueEntry()
        issue_entry.update_from_dict(self)
        issue_entry.issue_timestamp = issue.issue_timestamp
        issue_entry.result_timestamp = "-"
        issue_entry.rtype = "-"
        issue_entry.rcode = "-"
        issue_entry.iid = str(issue.iid)
        issue_entry.itype = issue.itype.name
        issue_entry.ename = issue.ename
        issue_entry.emsg = issue.emsg
        issue_entry.etrace = issue.etrace
        if self.iid == "-":
            self.iid = issue.iid
        self._issues.append(issue_entry)

    def append_step_issue(self, sr, issue):
        issue_entry = IssueEntry()
        issue_entry.update_from_dict(self)
        issue_entry.update_from_dict(sr)
        issue_entry.issue_timestamp = issue.issue_timestamp
        issue_entry.result_timestamp = "-"
        issue_entry.rtype = "-"
        issue_entry.rcode = "-"
        issue_entry.iid = str(issue.iid)
        issue_entry.itype = issue.itype.name
        issue_entry.ename = issue.ename
        issue_entry.emessage = issue.emsg
        issue_entry.etrace = issue.etrace
        if self.iid == "-":
            self.iid = issue.iid
        self._issues.append(issue_entry)

    def get_reportable(self):
        class Reportable:
            def __init__(this):
                this.result = OrderedDict()
                this.result.update(self.items())
                this.steps = self._steps
            def has_steps(this):
                return len(this.steps) > 0

        return Reportable()

SessionResult = TestObjectResult
StageResult = TestObjectResult
GroupResult = TestObjectResult
ModuleResult = TestObjectResult
FunctionResult = TestObjectResult

class SteppedResult(TestObjectResult):

    def __init__(self, test_obj):
        super().__init__(test_obj)

    def is_stepped(self):
        return True

    def get_step_type(self):
        pass

    def __create_step_result(self, step):
        sr = StepResult()
        sr.otype = step.otype
        sr.step_rtype = step.rtype == "-" and step.rtype or step.rtype.name
        sr.iid = step.iid
        sr.step_timestamp = step.timestamp
        sr.step_exec_context = step.exec_context.name
        sr.step_type = step.step_type.name
        sr.step_id = step.step_num
        sr.step_source = step.source
        sr.purpose = step.purpose
        sr.observation = step.observation
        sr.expectation = step.expectation
        sr.assert_message = step.assert_message
        sr.step_props = str(step.step_props)
        if step.has_issue():
            self.append_step_issue(sr, step.issue)
        self._steps.append(sr)

    def set_steps(self, step_list):
        for step in step_list:
            if self.rtype != ResultTypeEnum.FAIL.name:
                if step.failed():
                    self.rtype = ResultTypeEnum.FAIL.name
                    self.rcode = "{}_STEP_FAILURE".format(self.otype.upper())
                    self.iid = step.iid
                elif step.erred():
                    self.rtype = ResultTypeEnum.ERROR.name
                    self.rcode = "{}_STEP_ERROR".format(self.otype.upper())
                    self.iid = step.iid
            self.__create_step_result(step)

class FixtureResult(SteppedResult):

    def __init__(self, test_obj, fixture):
        super().__init__(test_obj)
        self.otype = TestObjectTypeEnum.Fixture.name
        self.fixture = fixture.qname
        self.fixtype = fixture.type.name
        self.evars = str(fixture.tvars.evars)
        self.data_record = "-"

    def set_steps(self, step_list):
        if not step_list:
            self.rtype = "PASS"
            self.rcode = "NO_STEPS_RECORDED"
            return

        super().set_steps(step_list)

class TestResult(SteppedResult):

    def __init__(self, test_obj):
        super().__init__(test_obj)
        self.otype = TestObjectTypeEnum.Test.name
        if not test_obj.tvars.data.record.is_empty():
            self.data_record = test_obj.tvars.data.record

    def set_steps(self, step_list):
        if not step_list:
            self.rtype = "NOT_A_TEST"
            self.rcode = "NO_STEPS_RECORDED"
            return

        super().set_steps(step_list)

# class Issue(DictWrapper):
#
#     def __init__(self, test_obj, *, iid, itype, istype, ename, emessage, trace):
#         super().__init__(IssueEntryEnum)
#         base = Info("-", test_obj)
#         self.update_from_dict(base.items())
#         self.issue_timestamp = datetime.datetime.now().timestamp()
#         self.iid = iid
#         self.itype = itype.name
#         self.istype = istype.name
#         self.ename = ename
#         self.emessage = emessage
#         self.trace = trace


import os
import copy
import re
import datetime
import platform
from arjuna.lib.reader.hocon import HoconFileReader, HoconStringReader, HoconConfigDictReader

from arjuna.tpi.enums import ArjunaOption
from .validator import ConfigValidator
from .config import Config, SetuConfig, UserConfig
from pprint import pprint

class ConfigCreator:
    SETU_CONF_DESC_MAP = None

    @classmethod
    def init(cls):
        if cls.SETU_CONF_DESC_MAP is not None:
            return
        my_dir = os.path.dirname(os.path.realpath(__file__))
        setu_conf_desc_file = os.path.join(my_dir, "res", "arjuna_conf_desc.conf")
        cls.__process_setu_conf_desc(setu_conf_desc_file)  

    @classmethod
    def __process_setu_conf_desc(cls, type_path):
        cls.SETU_CONF_DESC_MAP = cls.get_flat_map_from_hocon_string_for_setu_types(
            HoconFileReader(type_path)
        )    

    @classmethod
    def __setu_conf_key(cls, key):
        try:
            key = re.sub(r"\"(.*?)\"", r"\1", key)
            return ArjunaOption[key]
        except Exception:
            raise Exception("Config option <{}> is not a valid ArjunaOption constant".format(key))

    @classmethod
    def __get_flat_map_from_hocon_string(cls, hreader):
        hreader.process()
        return {i.upper().strip().replace(".", "_"): j for i, j in hreader.get_flat_map().items()}

    @classmethod      
    def get_flat_map_from_hocon_string_for_setu_types(cls, hreader):
        return {cls.__setu_conf_key(k):v for k,v in cls.__get_flat_map_from_hocon_string(hreader).items()}

    @classmethod
    def create_config_for_raw_map(self, map, validator_func):
        out_map = {}
        for conf_name, raw_value in map.items():
            validator_name, validator = validator_func(conf_name)
            try:
                is_not_set = False
                try:
                    is_not_set = raw_value.lower() == "not_set"
                except:
                    pass
                finally:
                    if is_not_set:
                        out_map[conf_name] = "not_set"
                    else:
                        out_map[conf_name] = validator(raw_value)
            except:
                raise Exception("Config option value <{}>(type:{}) for <{}> option did not pass the validation check: [{}]".format(
                    raw_value, type(raw_value), conf_name, validator_name)
                )
        return out_map

    @classmethod
    def create_conf(cls, processor, setu_conf, user_conf):
        config = Config()
        config.setu_config = SetuConfig(setu_conf)
        config.user_config = UserConfig(user_conf)
        config.processor = processor
        return config

    @classmethod
    def __create_new_conf(cls, processor, source, setu_dict=None, user_dict=None):
        out_setu = source and copy.deepcopy(source.setu_config.as_map()) or {}
        out_user = source and copy.deepcopy(source.user_config.as_map()) or {}
        if setu_dict:
            out_setu.update(setu_dict)
        if user_dict:
            out_user.update(user_dict)
        return cls.create_conf(processor, out_setu, out_user)

    @classmethod
    def create_new_conf(cls, processor, source, cdict):
        if not cdict:
            return cls.__create_new_conf(processor, source)

        custom_setu_conf = None
        if "arjunaOptions" in cdict:
            d = HoconConfigDictReader(cdict["arjunaOptions"])
            custom_raw_setu_config_map = cls.get_flat_map_from_hocon_string_for_setu_types(d)
            custom_setu_conf = cls.create_config_for_raw_map(custom_raw_setu_config_map, processor.get_setu_option_validator)

        custom_user_conf = None
        if "userOptions" in cdict:
            d = HoconConfigDictReader(cdict["userOptions"])
            project_raw_user_config_map = cls.__get_flat_map_from_hocon_string(d)
            custom_user_conf = cls.create_config_for_raw_map(
                project_raw_user_config_map, 
                processor.get_user_option_validator
            )
        return cls.__create_new_conf(processor, source, custom_setu_conf, custom_user_conf) 

class BaseConfigProcessor:

    def __init__(self):
        self.__setu_type_map = {}
        self.__config = None

    @property
    def config(self):
        return self.__config

    @config.setter
    def _config(self, config):
        self.__config = config

    def pass_through(self, input):
        return input

    def get_setu_option_validator(self, conf_name):
        validator_name = ConfigCreator.SETU_CONF_DESC_MAP[conf_name]
        return validator_name, getattr(ConfigValidator, validator_name.lower())

    def get_user_option_validator(self, conf_name):
        return "pass_through", self.pass_through


class CentralConfigLoader(BaseConfigProcessor):
    OS_MAP = {
        'Windows': 'windows',
        'Darwin': 'mac',
        'Linux': 'linux'
    }

    def __init__(self, project_root_dir, runid=None):
        self.__my_dir = os.path.dirname(os.path.realpath(__file__))
        self.__setu_central_confg_file = os.path.join(self.__my_dir, "res", "arjuna.conf")
        self.__project_root_dir = project_root_dir
        self.__project_name = os.path.basename(self.__project_root_dir)
        self.__runid = runid and runid or "mrun"
        self.__process()

    def __process(self):
        # Processes central conf based on root directory
        f = open(self.__setu_central_confg_file, "r")
        contents = f.read()
        f.close()
        arjuna_root = os.path.abspath(os.path.join(self.__my_dir, "..", ".."))

        irunid = "{}-{}".format(datetime.datetime.now().strftime("%Y.%m.%d-%H.%M.%S.%f")[:-3], self.__runid)
        test_module_import_prefix = "{}.tests.modules.".format(self.__project_name)
        conf_fixtures_import_prefix = "{}.fixtures.".format(self.__project_name)

        contents = contents.replace("<ARJUNA_ROOT_DIR>", arjuna_root)
        contents = contents.replace("<PROJECT_ROOT_DIR>", self.__project_root_dir)
        contents = contents.replace("<PROJECT_NAME>", self.__project_name)
        contents = contents.replace("<IRUNID>", irunid)
        contents = contents.replace("<TEST_MODULE_IMPORT_PREFIX>", test_module_import_prefix)
        contents = contents.replace("<FIXTURES_IMPORT_PREFIX>", conf_fixtures_import_prefix)
        contents = contents.replace("<HOST_OS>", CentralConfigLoader.OS_MAP[platform.system()])
        raw_config_map = ConfigCreator.get_flat_map_from_hocon_string_for_setu_types(
            HoconStringReader(contents)
        )
        self._config = ConfigCreator.create_conf(
            self,
            ConfigCreator.create_config_for_raw_map(raw_config_map, self.get_setu_option_validator), 
            {}
        )


class ProjectConfigCreator(BaseConfigProcessor):

    def __init__(self, central_conf):
        self.__central_conf = central_conf
        self.__process()

    def __process(self):
        # Process project conf
        project_conf_file = self.__central_conf.setu_config.value(ArjunaOption.PROJECT_CONF_FILE)
        project_hreader = HoconFileReader(project_conf_file)
        project_hreader.process()
        cdict = project_hreader.get_map()

        self._config = ConfigCreator.create_new_conf(self, self.__central_conf, cdict)






from arjuna.tpi.enums import ArjunaOption

from .guiautomator import GuiAutomator
from .handler import Handler
from enum import Enum, auto

class MobileView(Enum):
    NATIVE_APP = auto()
    WEBVIEW = auto()


class ViewContextHandler(Handler):

    def __init__(self, automator: GuiAutomator):
        super().__init__(automator)

    def switch_to_view_context(self, view_name):
        self._act(TestAutomatorActionBodyCreator.switch_to_view_context(view_name))

    def get_current_view_context(self):
        response = self._act(TestAutomatorActionBodyCreator.get_current_view_context())
        return response["data"]["viewContext"]

    def get_all_view_contexts(self):
        response = self._act(TestAutomatorActionBodyCreator.get_all_view_contexts())
        return response["data"]["viewContexts"]

    def switch_to_native_view(self):
        self.switch_to_view_context(MobileView.NATIVE_APP.name)

    def switch_to_web_view(self, pkg):
        self.switch_to_view_context(
            "{}_{}".format(MobileView.WEBVIEW.name, pkg)
        )

    def switch_to_first_web_view(self):
        for context_name in self.get_all_view_contexts():
            if context_name.find(MobileView.WEBVIEW.name) != -1:
                self.switch_to_view_context(context_name)
                break

    def _does_name_represent_web_view(self, view_name):
        return view_name.lower().find(MobileView.WEBVIEW.name.lower()) != -1 

    def is_web_view(self):
        return self._does_name_represent_web_view(self.get_current_view_context())

    def is_native_view(self):
        return self.get_current_view_context().lower() == MobileView.WEBVIEW.name.lower()


import json
from arjuna.setu.config.processor import ConfigCreator, CentralConfigLoader, ProjectConfigCreator
from arjuna.tpi.enums import ArjunaOption
from arjuna.lib.reader.hocon import HoconStringReader, HoconConfigDictReader

from arjuna.setu.types import SetuManagedObject
from arjuna.setuext.data_broker.databroker import DataBroker


class TestConfigurator:

    def __init__(self):
        self.__default_ref_config = None
        self.__config_map = {}
        self.__cli_central_config = None
        self.__cli_test_config = None

    def __create_config_from_option_dicts(self, reference, arjuna_options, user_options):
        crawdict = {
            "arjunaOptions": arjuna_options,
            "userOptions": user_options
        }
        hreader = HoconStringReader(json.dumps(crawdict))
        hreader.process()
        config = ConfigCreator.create_new_conf(
            self.__default_ref_config.processor,
            reference,
            hreader.get_map()
        )
        return config

    def __init_cli_dicts(self, arjunaCentralOptions={}, arjunaTestOptions={}, userCentralOptions={}, userTestOptions={}):
        self.__cli_central_config = self.__create_config_from_option_dicts(None, arjunaCentralOptions, userCentralOptions)
        self.__cli_test_config = self.__create_config_from_option_dicts(None, arjunaTestOptions,
                                                                        userTestOptions)

    def init(self, root_dir, cli_config):
        self.__default_ref_config = CentralConfigLoader(root_dir).config
        self.__config_map[self.__default_ref_config.setu_id] = self.__default_ref_config
        cli_config = cli_config and cli_config or {}
        self.__init_cli_dicts(**cli_config)
        return self.__default_ref_config.setu_id

    def create_project_conf(self):
        project_conf_loader = ProjectConfigCreator(self.__default_ref_config)
        self.__default_ref_config = project_conf_loader.config
        self.__default_ref_config.process_arjuna_options()
        self.__default_ref_config.update(self.__cli_central_config)
        self.__config_map[self.__default_ref_config.setu_id] = self.__default_ref_config
        return self.__default_ref_config

    def get_central_config(self):
        return self.__default_ref_config

    def get_config(self, setu_id):
        return self.__config_map[setu_id]

    def get_arjuna_option_value(self, config_setu_id, option):
        sname = ArjunaOption[option.upper().strip().replace(".", "_")]
        rvalue = self.__config_map[config_setu_id].setu_config.value(sname)
        return rvalue

    def get_user_option_value(self, config_setu_id, option):
        sname = option.upper().strip().replace(".", "_")
        rvalue = self.__config_map[config_setu_id].user_config.value(sname)
        return rvalue

    def get_central_arjuna_option_value(self, option):
        sname = ArjunaOption[option.upper().strip().replace(".", "_")]
        rvalue = self.__default_ref_config.setu_config.value(sname)
        return rvalue

    def get_central_user_option_value(self, option):
        sname = option.upper().strip().replace(".", "_")
        rvalue = self.__default_ref_config.user_config.value(sname)
        return rvalue

    def register_config(self, arjuna_options, has_parent, user_options, parent_config_id):
        reference = has_parent and self.__config_map[parent_config_id] or self.__default_ref_config
        config = self.__create_config_from_option_dicts(reference, arjuna_options, user_options)
        config.update(self.__cli_test_config)
        config.process_arjuna_options()
        self.__config_map[config.setu_id] = config
        return config


class TestSession(SetuManagedObject):

    def __init__(self):
        super().__init__()
        self.__configurator = TestConfigurator()
        self.__data_broker = DataBroker()

    @property
    def configurator(self):
        return self.__configurator

    @property
    def data_broker(self):
        return self.__data_broker

    def init(self, root_dir, cli_config):
        return self.__configurator.init(root_dir, cli_config)


import os
from setuptools import setup, find_packages

# my_dir =  os.path.dirname(os.path.realpath(__file__))

# dynamic_dirs = [
#     os.path.join(my_dir, "arjuna", p) for p in (
#         "res",
#         "third_party"
#     )
# ]

packages = find_packages()
# packages.extend(dynamic_dirs)

# print(packages)

this_directory = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "arjuna",
    version = "0.7.7",
    url = "https://testmile.com/arjuna",
    description = "Arjuna is a Python based test automation framework developed by Rahul Verma (www.rahulverma.net).",
    author = "Rahul Verma",
    author_email = "rv@testmile.com",
    packages = packages,
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data = {
        '' :  [
                    "*.txt",
                    "*.md",
                    "*.cfg",
        ],
        'arjuna' : [
                    "res/st/*.xml", 
                    "res/*.help", 
                    "res/*.txt", 
                    "res/scripts/arjuna_launcher.py",
                    "setu/config/res/*.conf", 
                    "setu/config/res/*.txt"
                ]
    },
    install_requires = ["flask", "flask-RESTful", "waitress", "requests", "selenium", "xlrd", "xlwt", "pyparsing", "pyhocon"],
    keywords = "arjuna setu unitee selenium testing automation page-object",
    license = "Apache License, Version 2.0",
    classifiers=[
    'Environment :: Console',
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Topic :: Software Development :: Quality Assurance',
    'Topic :: Software Development :: Testing',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: Implementation :: CPython',
    'Operating System :: OS Independent',
    'Natural Language :: English'
    ]
)


from .guiautomator import GuiAutomatorHandler

class Receiver:
    AUTOMATOR_HANDLERS = dict()

    @classmethod
    def create_gui_automator_dispatcher(cls, setu_id):
        handler = GuiAutomatorHandler(setu_id)
        cls.AUTOMATOR_HANDLERS[setu_id] = handler
        return handler

    @classmethod
    def create_gui_element_dispatcher(cls, automator_setu_id, element_setu_id):
        handler = cls.AUTOMATOR_HANDLERS[automator_setu_id]
        return handler.create_gui_element_handler(element_setu_id)


    


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import importlib

from arjuna.tpi.enums import ArjunaOption
from arjuna.unitee.test.objects.fixture import Fixture
from arjuna.unitee.enums import FixtureTypeEnum
from arjuna.unitee.loader import kfactory
from arjuna.unitee.enums import *

class _FixtureDef:
    def __init__(self, ftype, fix_func):
        self.__func = fix_func
        self.__type = ftype

    @property
    def name(self):
        return self.__func.name

    @property
    def func(self):
        return self.__func.obj

    @property
    def qname(self):
        return self.__func.qname

    @property
    def type(self):
        return self.__type

    def create_fixture(self):
        return Fixture(self)

class FixturesDef:
    def __init__(self):
        self.__fixtures = {
            FixtureTypeEnum.INIT_SESSION : None,
            FixtureTypeEnum.END_SESSION : None,
            FixtureTypeEnum.INIT_EACH_STAGE : None,
            FixtureTypeEnum.END_EACH_STAGE : None,
            FixtureTypeEnum.INIT_STAGE : None,
            FixtureTypeEnum.END_STAGE : None,
            FixtureTypeEnum.INIT_EACH_GROUP : None,
            FixtureTypeEnum.END_EACH_GROUP : None,
            FixtureTypeEnum.INIT_GROUP : None,
            FixtureTypeEnum.END_GROUP : None,
            FixtureTypeEnum.INIT_EACH_MODULE : None,
            FixtureTypeEnum.END_EACH_MODULE : None,
            FixtureTypeEnum.INIT_MODULE: None,
            FixtureTypeEnum.END_MODULE: None,
            FixtureTypeEnum.INIT_EACH_FUNCTION: None,
            FixtureTypeEnum.END_EACH_FUNCTION: None,
            FixtureTypeEnum.INIT_EACH_TEST: None,
            FixtureTypeEnum.END_EACH_TEST: None
        }

    def add_fixture_func(self, f_type, f_func):
        self.__fixtures[f_type] = _FixtureDef(f_type, f_func)

    def build(self, f_type):
        if self.__fixtures[f_type]:
            return self.__fixtures[f_type].create_fixture()
        else:
            return None

    def get_fixture_name(self, ftype):
        return self.__fixtures[ftype] and self.__fixtures[ftype].name or None

class ConfiguredFixtureHelper:

    @staticmethod
    def configure_fixture(fixdef, ftypestr, mname, fname):
        from arjuna.tpi import Arjuna
        fix_prefix = Arjuna.get_central_config().get_arjuna_option_value(ArjunaOption.UNITEE_PROJECT_FIXTURES_IMPORT_PREFIX).as_string()
        module = importlib.import_module(fix_prefix + mname)
        func = getattr(module, fname)
        fixdef.add_fixture_func(FixtureTypeEnum[ftypestr.upper()], kfactory.create_fixture(ftypestr, func))


import abc
import types

class Handler(metaclass=abc.ABCMeta):

    def __init__(self, element_obj):
        self.__element = element_obj
        self.__config = element_obj.config

    @property
    def element(self):
        return self.__element

    @property
    def config(self):
        return self.__config

    def _act(self, json_dict):
        return self.__element.actor_callable(json_dict)


from arjuna.unitee.types.containers import *

def get_children_tags(xml_node):
    return [child.tag.lower() for child in list(xml_node)]

def has_child(xml_node, child_tag):
    return child_tag.lower() in get_children_tags(xml_node)

def convert_to_cidict(xml_node):
    return CIStringDict({child.tag:child for child in list(xml_node)})

def convert_attribs_to_cidict(xml_node):
    return CIStringDict({attr: value for attr, value in xml_node.attrib.items()})


from arjuna.setu.types import SetuManagedObject
from arjuna.setuext.guiauto.impl.element.guielement import GuiElement
from arjuna.setuext.guiauto.impl.locator.emd import SimpleGuiElementMetaData

# UUID is for client reference. Agent does not know about this.
class GuiWebSelect(SetuManagedObject):

    def __init__(self, automator, locator_meta_data, parent=None):
        super().__init__()
        self._wrapped_main_element = automator.create_element(locator_meta_data)
        tag = self._wrapped_main_element.get_tag_name()
        if tag.lower() != "select":
            raise Exception("The element should have a 'select' tag for WebSelect element. Found: " + tag)
        self._multi = self.__is_multi_select()
        self.__options = self._wrapped_main_element.create_multielement(
            SimpleGuiElementMetaData("tag_name", "option")
        )

    def __is_multi_select(self):
        return self._wrapped_main_element.get_attr_value("multiple", optional=True) is True or self._wrapped_main_element.get_attr_value("multi", optional=True) is True

    def is_multi_select(self):
        return self._multi

    def has_index_selected(self, index):
        return self.__options.get_instance_at_index(index).is_selected()

    def has_value_selected(self, value):
        return self.__options.get_instance_by_value(value).is_selected()

    def has_visible_text_selected(self, text):
        return self.__options.get_instance_by_visible_text(text).is_selected()

    def get_first_selected_option_text(self):
        option = self.__options.get_first_selected_instance()
        return option.get_text_content()

    def select_by_index(self, index):
        self.__options.get_instance_at_index(index).select()

    def select_by_ordinal(self, ordinal):
        return self.select_by_index(ordinal-1)

    def select_by_visible_text(self, text):
        return self.__options.get_instance_by_visible_text(text).select()

    def select_by_value(self, value):
        return self.__options.get_instance_by_value(value).select()

    # The following methods deal with multi-select and would be implemented later.

    def __validate_multi_select(self):
        if not self.is_multi_select():
            raise Exception("Deselect actions are allowed only for a multi-select dropdown.")

    def deselect_by_value(self, value):
        self.__validate_multi_select()
        return self.__options.get_instance_by_value(value).deselect()

    def deselect_by_index(self, index):
        pass

    def deselect_by_visible_text(self, text):
        pass

    def get_selected_options(self):
        pass

    def are_visible_texts_selected(self, text_list):
        pass

    def are_values_selected(self, text_list):
        pass

    def all_options(self):
        pass

    def select_by_values(self, value_list):
        pass

    def deselect_by_values(self, value_list):
        pass

    def select_by_indices(self, indices):
        pass

    def deselect_by_indices(self, indices):
        pass

    def select_by_visible_texts(self, text_list):
        pass

    def deselect_by_visible_texts(self, text_list):
        pass


import os

from arjuna.tpi.enums import ArjunaOption
from .nsloader import GuiNamespaceLoaderFactory
from arjuna.setu.types import SetuConfiguredObject
from arjuna.setuext.guiauto.impl.locator.emd import SimpleGuiElementMetaData, GuiElementMetaData, Locator


class Gui(SetuConfiguredObject):

    def __init__(self, name_store, namespace_dir, automator, label, file_def_path):
        super().__init__(automator.config)
        self.__name_store = name_store
        self.__namespace_dir = namespace_dir
        self.__automator = automator
        self.__auto_context = self.config.setu_config.get_guiauto_context()
        self.__file_def_path = os.path.abspath(os.path.join(self.__namespace_dir, file_def_path.strip()))
        self.__ns = None
        ns_name = "file_ns::" + self.__file_def_path.lower()
        if name_store.has_namespace(ns_name):
            self.__ns = name_store.get_namespace(ns_name)
        else:
            self.__ns = name_store.load_namespace(
                ns_name, 
                GuiNamespaceLoaderFactory.create_namespace_loader(self.__file_def_path)
        )

        self.__children = []

    def add_child(self, label, automator, file_def_path):
        self.__children.append(
            Gui(self.__name_store, self.__namespace_dir, label, self.__automator, file_def_path)
    )

    def get_emd(self, locators):
        final_locators = []
        for raw_locator in locators:
            if raw_locator["withType"].upper().strip() == "ASSIGNED_NAME":
                emd = self.__ns.get_meta_data(raw_locator["withValue"], self.__auto_context)
                final_locators.extend(emd.raw_locators)
            else:
                final_locators.append(Locator(ltype=raw_locator["withType"], lvalue=raw_locator["withValue"]))
        return GuiElementMetaData(final_locators)

    def create_dispatcher(self):
        # Pages don't use any dispatcher
        pass

class GuiFactory:

    @classmethod
    def create_app_from_dir(cls, name, automator, app_def_dir):
        considered_path = app_def_dir
        if not os.path.isdir(considered_path):
            ns_dir = automator.config.value(ArjunaOption.GUIAUTO_NAMESPACE_DIR)
            full_path = os.path.join(ns_dir, considered_path)
            considered_path = os.path.abspath(full_path)
            if not os.path.isdir(considered_path):
                raise Exception("Provided root definition path is not a directory: {}".format(app_def_dir))

        app = Gui(automator, os.path.join(considered_path, "HomePage.gns"), label=name)
        children_dir = os.path.join(considered_path, "children")
        if os.path.isdir(children_dir):
            lfiles = os.listdir(children_dir)
            for f in lfiles:
                cpath = os.path.join(children_dir, f)
                if os.path.isfile(cpath):
                    base_name = os.path.basename(cpath)
                    app.add_child(base_name, cpath)

    @classmethod
    def create_gui(cls, automator, def_path):
        return Gui(automator, def_path)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import copy
import os
import codecs
import sys
import io
# sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding="UTF-8")
# sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding="UTF-8")

import logging
from arjuna.lib.utils import thread_utils
from arjuna.client.testsession import DefaultTestSession
from arjuna.lib.thread.decorators import *
from arjuna.lib.utils import sys_utils
from arjuna.lib.adv.proxy import ROProxy
from arjuna.lib.adv.decorators import singleton
from arjuna.tpi.enums import ArjunaOption
import codecs
import sys


@singleton
class ArjunaSingleton:

    def __init__(self):
        self.__project_root_dir = None
        self.__test_session = None
        self.__central_config = None
        self.__test_context_config_map = dict()
        self.__run_id = None
        self.__unitee = None

        # From central config
        self.prop_enum_to_prop_path_map = {}
        # It is base config object
        self.arjuna_options = {}
        # NAme of thread: Base configuration
        self.thread_map = {}
        # DefaultStringKeyValueContainer
        self.exec_var_map = {}
        # DefaultStringKeyValueContainer
        self.user_options = {}

        # Check relevance of above and following
        self.prog = "Arjuna"

        self.dl = None
        self.log_file_discovery_info = False
        self._data_references = {}
        self.central_conf = None
        self.__console = None
        self.__logger = None

    def init(self, project_root_dir, cli_config, run_id):
        self.__project_root_dir = project_root_dir
        self.__test_session = DefaultTestSession()
        self.__central_config = self.__test_session.init(project_root_dir, cli_config, run_id)

        self.__load_console()

        from arjuna.unitee import Unitee
        project_name = self.__central_config.get_arjuna_option_value(ArjunaOption.PROJECT_NAME).as_string()
        self.__unitee = Unitee(self.__test_session, self.__central_config)

        return self.__central_config

    def get_logger(self):
        return self.__logger

    def get_console(self):
        return self.__console

    @classmethod
    def get_test_session(cls):
        return cls.__SESSION

    @classmethod
    def has_configuration(cls, config_name):
        return config_name.upper() in cls.thread_map

    @classmethod
    def has_property(cls, config_name, path):
        return cls.has_configuration(config_name) and path.upper() in cls.thread_map[config_name]

    @classmethod
    def merge_configuration(cls, source_config_name, target_config_name):
        if cls.has_configuration(source_config_name):
            if cls.has_configuration(target_config_name):
                cls.thread_map[target_config_name.upper()] = cls.thread_map[source_config_name].clone()
            else:
                cls.register_new_configuration(target_config_name.upper())
                cls.thread_map[target_config_name.upper()] = cls.thread_map[source_config_name.upper()].clone()
        else:
            raise Exception("No source configuration found for name: %s".format(source_config_name))

    @classmethod
    def register_new_configuration(cls, config_name):
        cls.thread_map[config_name.upper()] = {}

    @classmethod
    def update_thread_config(cls, config_name, config):
        cls.thread_map[config_name].clone_add(config)

    @classmethod
    def value(cls, prop_name):
        # uc_prop_name = None
        # if type(prop_name) is not str:
        #     prop_name = self.__get_prop_path_for_enum(prop_name).upper()
        uc_prop_name = prop_name.upper()
        tname = thread_utils.get_current_thread_name()
        if tname in cls.thread_map and uc_prop_name in cls.thread_map[tname]:
            return cls.thread_map[tname][uc_prop_name].value
        elif uc_prop_name in cls.arjuna_options:
            return cls.arjuna_options[uc_prop_name].value
        else:
            raise Exception("No property configred for name: {}".format(prop_name))

    @classmethod
    def get_configured_name(self, section_name, internal_name):
        return self.strings_manager.get_configured_name(section_name, internal_name)

    @classmethod
    def get_problem_text(self, problem_code):
        return self.strings_manager.get_problem_text(problem_code)

    @classmethod
    def get_info_message_text(self, code):
        return self.strings_manager.get_info_message_text(code)

    @classmethod
    def clone_evars(self):
        return copy.deepcopy(self.exec_var_map)

    @classmethod
    def clone_user_options(self):
        return copy.deepcopy(self.user_options)

    def __load_console(self):
        dl = logging.getLevelName(self.__central_config.get_arjuna_option_value(ArjunaOption.LOG_CONSOLE_LEVEL).as_string().upper())
        log_dir = self.__central_config.get_arjuna_option_value(ArjunaOption.LOG_DIR).as_string()
        if not os.path.isdir(log_dir):
            os.makedirs(log_dir)
        fl = logging.getLevelName(self.__central_config.get_arjuna_option_value(ArjunaOption.LOG_FILE_LEVEL).as_string().upper())
        fname = self.__central_config.get_arjuna_option_value(ArjunaOption.PYTHON_LOG_NAME).as_string()
        lpath = os.path.join(log_dir, fname)

        logger = logging.getLogger(self.prog)
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(dl)
        fh = logging.FileHandler(lpath, "w", 'utf-8')
        fh.setLevel(fl)
        f_fmt = logging.Formatter(u'[%(levelname)5s]\t%(asctime)s\t%(pathname)s::%(module)s.%(funcName)s:%(lineno)d\t%(message)s')
        c_fmt = logging.Formatter(u'[%(levelname)5s]\t%(message)s')
        ch.setFormatter(c_fmt)
        fh.setFormatter(f_fmt)
        logger.addHandler(ch)
        logger.addHandler(fh)

        class __console:
            lock = threading.RLock()

            def __init__(self):
                # self.lock = threading.RLock()
                self.separator = os.linesep
                self.log_display_level = dl
                self.logger = logger

            @sync_method('lock')
            def __log(self, message, err=False):
                if err:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.error(mpart)
                else:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.info(mpart)

            @sync_method('lock')
            def __print(self, message):
                print(message, end='', file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprint(self, message):
                print(message, end='', file=sys.stderr, flush=True)

            @sync_method('lock')
            def __println(self, message):
                print(message, file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprintln(self, message):
                print(message, file=sys.stderr, flush=True)

            @sync_method('lock')
            def __msg(self, *messages):
                return " ".join([str(m) for m in messages])

            @sync_method('lock')
            def display(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__println(message)

            @sync_method('lock')
            def display_error(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message, err=True)
                if should_print:
                    self.__eprintln(message)

            @sync_method('lock')
            def error_for_console(self, *messages):
                message = self.__msg(*messages)
                self.__eprintln(message)

            @sync_method('lock')
            def display_on_same_line(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__print(message)

            @sync_method('lock')
            def marker(self, length, symbol='-'):
                self.display(symbol * length)

            @sync_method('lock')
            def marker_error(self, length, symbol='-'):
                self.display_error(symbol * length)

            @sync_method('lock')
            def marker_on_same_line(self, length=40):
                self.display_on_same_line("-" * length)

            @sync_method('lock')
            def display_key_value(self, key, value):
                message = "%s %s".format(key, value)
                self.display(message)

            @sync_method('lock')
            def __get_formatted_key_value(self, key, value, left_padding):
                if not left_padding:
                    message = "| {:20s}| {}".format(key, value)
                else:
                    message = "| {}| {}".format(key.ljust(left_padding), value)
                return message

            @sync_method('lock')
            def __display_paddedKV(self, key, value, print_func, left_padding):
                print_func(self.__get_formatted_key_value(key, value, left_padding))

            @sync_method('lock')
            def display_padded_key_value(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display, left_padding)

            @sync_method('lock')
            def display_padded_key_value_error(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display_error, left_padding)

            @sync_method('lock')
            def display_exception_block(self, e, strace):
                self.marker_error(80)
                self.display_padded_key_value_error("Exception Type", e.__class__.__name__, 30)
                self.display_padded_key_value_error("Exception Message", str(e), 30)
                self.display_padded_key_value_exception_trace("Exception Trace", strace, 30)
                self.marker_error(80)

            def set_central_log_level(self, level):
                self.central_log_level = level

            @sync_method('lock')
            def display_multiline_key_value(self, key, value, left_padding=30):
                value = str(value)
                ctrace_parts = value.replace("\t", " ").replace("\r\n\r\n","\r\n").replace("\n\n","\n").split(sys_utils.get_line_separator())

                header = self.__get_formatted_key_value(key, ctrace_parts[0], left_padding)
                self.__log(header)
                # if should_print:
                #     self.error_for_console(header)

                for s in ctrace_parts[1:]:
                    message = self.__get_formatted_key_value("", s, left_padding)
                    self.__log(message)
                    # if should_print:
                    #     self.(message)

            display_padded_key_value_exception_trace = display_multiline_key_value
        self.__logger = logger
        self.__console = __console()

    def get_unitee_instance(self):
        return self.__unitee

    def get_central_config(self):
        return self.__central_config

    def create_gui_automator(self):
        from arjuna.client.guiauto.automator import DefaultGuiAutomator
        return DefaultGuiAutomator(self.__central_config)


class Arjuna:
    ARJUNA_SINGLETON = None

    @classmethod
    def init(cls, project_root_dir, cli_config=None, run_id=None):
        cls.ARJUNA_SINGLETON = ArjunaSingleton()
        cls.ARJUNA_SINGLETON.init(project_root_dir, cli_config, run_id)

    @classmethod
    def get_logger(cls):
        return cls.ARJUNA_SINGLETON.get_logger()

    @classmethod
    def get_console(cls):
        return cls.ARJUNA_SINGLETON.get_console()

    @classmethod
    def get_central_config(cls):
        return cls.ARJUNA_SINGLETON.get_central_config()

    @classmethod
    def get_unitee_instance(cls):
        return cls.ARJUNA_SINGLETON.get_unitee_instance()

    @classmethod
    def create_gui_automator(cls):
        return cls.ARJUNA_SINGLETON.create_gui_automator()

    @classmethod
    def get_central_arjuna_option(cls, option):
        return cls.get_central_config().get_arjuna_option_value(option)

    @classmethod
    def get_central_user_option(cls, option):
        return cls.get_central_config().get_user_option_value(option)


from arjuna.client.core.connector import BaseSetuObject, SetuArg
from arjuna.client.core.action import *
from arjuna.client.core.config import ArjunaComponent


class GuiAutoComponentFactory:

    @staticmethod
    def Element(session, automator, setuId):
        return DefaultGuiElement(session, automator, setuId)

    @staticmethod
    def MultiElement(session, automator, setuId):
        return DefaultGuiMultiElement(session, automator, setuId)

    @staticmethod
    def DropDown(session, automator, setuId):
        return DefaultDropDown(session, automator, setuId)

    @staticmethod
    def RadioGroup(session, automator, setuId):
        return DefaultRadioGroup(session, automator, setuId)

    @staticmethod
    def Alert(session, automator, setuId):
        return DefaultAlert(session, automator, setuId)

    @staticmethod
    def MainWindow(session, automator, setuId):
        return DefaultMainWindow(session, automator, setuId)

    @staticmethod
    def DomRoot(session, automator):
        return DefaultDomRoot(session, automator)

    @staticmethod
    def Browser(session, automator):
        return DefaultBrowser(session, automator)


class BaseComponent(BaseSetuObject):

    def __init__(self, test_session, app_automator):
        super().__init__()
        self.__test_session = test_session
        self.__automator = app_automator

        if app_automator.is_gui():
            self._set_gui_setu_id_arg(app_automator.get_setu_id())
        else:
            self._set_automator_setu_id_arg(app_automator.get_setu_id())

        self._set_test_session_setu_id_arg(test_session.get_setu_id())

    def _get_automator(self):
        return self.__automator

    def _get_test_session(self):
        return self.__test_session


class BaseElement(BaseComponent):

    def __init__(self, test_session, app_automator, setu_id, index=None):
        super().__init__(test_session, app_automator)
        self._set_setu_id(setu_id)
        self._set_self_setu_id_arg("elementSetuId")

        if index is not None:
            self._add_args(
                SetuArg.arg("isInstanceAction", True),
                SetuArg.arg("instanceIndex", index)
            )


class DefaultGuiElement(BaseElement):

    def __init__(self, test_session, app_automator, setu_id, index=None):
        super().__init__(test_session, app_automator, setu_id, index)

    def enter_text(self, text):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ELEMENT_ENTER_TEXT, SetuArg.text_arg(text))

    def set_text(self, text):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ELEMENT_SET_TEXT, SetuArg.text_arg(text))

    def click(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ELEMENT_CLICK)

    def wait_until_clickable(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ELEMENT_WAIT_UNTIL_CLICKABLE)

    def check(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ELEMENT_CHECK)

    def uncheck(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ELEMENT_UNCHECK)


class DefaultGuiMultiElement(BaseElement):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)

    def IndexedElement(self, index):
        return DefaultGuiElement(self._get_test_session(), self._get_automator(), self.get_setu_id(), index)


class DefaultDropDown(BaseElement):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)

    def has_value_selected(self, value):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_HAS_VALUE_SELECTED, SetuArg.value_arg(value))
        return response.get_value_for_check_result()

    def has_index_selected(self, index):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_HAS_INDEX_SELECTED, SetuArg.index_arg(index))
        return response.get_value_for_check_result()

    def select_by_value(self, value):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_SELECT_BY_VALUE, SetuArg.value_arg(value))

    def select_by_index(self, index):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_SELECT_BY_INDEX, SetuArg.index_arg(index))

    def get_first_selected_option_value(self):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_GET_FIRST_SELECTED_OPTION_VALUE)
        return response.get_value_for_value_attr()

    def get_first_selected_option_text(self):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_GET_FIRST_SELECTED_OPTION_TEXT)
        return response.get_value_for_text()

    def has_visible_text_selected(self, text):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_HAS_VISIBLE_TEXT_SELECTED, SetuArg.text_arg(text))
        return response.get_value_for_check_result()

    def select_by_visible_text(self, text):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DROPDOWN_SELECT_BY_VISIBLE_TEXT, SetuArg.text_arg(text))


class DefaultRadioGroup(BaseElement):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)

    def has_value_selected(self, value):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.RADIOGROUP_HAS_VALUE_SELECTED, SetuArg.value_arg(value))
        return response.get_value_for_check_result()

    def has_index_selected(self, index):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.RADIOGROUP_HAS_INDEX_SELECTED, SetuArg.index_arg(index))
        return response.get_value_for_check_result()

    def select_by_value(self, value):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.RADIOGROUP_SELECT_BY_VALUE, SetuArg.value_arg(value))

    def select_by_index(self, index):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.RADIOGROUP_SELECT_BY_INDEX, SetuArg.index_arg(index))

    def get_first_selected_option_value(self):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.RADIOGROUP_GET_FIRST_SELECTED_OPTION_VALUE)
        return response.get_value_for_value_attr()


class DefaultAlert(BaseElement):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)

    def confirm(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ALERT_CONFIRM)

    def dismiss(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ALERT_DISMISS)

    def get_text(self):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ALERT_GET_TEXT)
        return response.get_value_for_text()

    def send_text(self, text):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.ALERT_SEND_TEXT, SetuArg.text_arg(text))


class DefaultBrowser(BaseComponent):

    def __init__(self, test_session, app_automator):
        super().__init__(test_session, app_automator)

    def go_to_url(self, url):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.BROWSER_GO_TO_URL, SetuArg.arg("url", url))

    def go_back(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.BROWSER_GO_BACK)

    def go_forward(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.BROWSER_GO_FORWARD)

    def refresh(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.BROWSER_REFRESH)


class DefaultFrame(BaseElement):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)
        self._set_self_setu_id_arg("elementSetuId")

    def focus(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.FRAME_FOCUS)

    def Frame(self, *withLocators):
        arg = [l.asMap() for l in withLocators]
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.FRAME_CREATE_FRAME, SetuArg.arg("locators", arg))
        return DefaultFrame(self._get_test_session(), self._get_automator(), response.get_value_for_element_setu_id())

    def ParentFrame(self):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.FRAME_GET_PARENT)
        return DefaultFrame(self._get_test_session(), self._get_automator(), response.get_value_for_element_setu_id())


class DefaultDomRoot(BaseComponent):

    def __init__(self, test_session, app_automator):
        super().__init__(test_session, app_automator)

    def focus(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DOMROOT_FOCUS)

    def Frame(self, *with_locators):
        arg = [l.as_map() for l in with_locators]
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.DOMROOT_CREATE_FRAME, SetuArg.arg("locators", arg))
        return DefaultFrame(self._get_test_session(), self._get_automator(), response.get_value_for_element_setu_id())

    def ParentFrame(self):
        raise Exception("DOM root does not have a parent frame.")


class AbstractBasicWindow(BaseElement):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)

    def get_title(self):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.WINDOW_GET_TITLE)
        return response.get_value_for_key("title").as_string()

    def focus(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.WINDOW_FOCUS)


class DefaultChildWindow(AbstractBasicWindow):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)

    def close(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.CHILD_WINDOW_CLOSE)

    def MainWindow(self):
        return self._get_automator().MainWindow()


class DefaultMainWindow(AbstractBasicWindow):

    def __init__(self, test_session, app_automator, setu_id):
        super().__init__(test_session, app_automator, setu_id)

    def maximize(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.MAIN_WINDOW_MAXIMIZE)

    def _take_element_finding_action(self, setu_action_type, *setu_args):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, setu_action_type, *setu_args)
        return response.get_value_for_element_setu_id()

    def ChildWindow(self, *withLocators):
        arg = [l.asMap() for l in withLocators]
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.MAIN_WINDOW_CREATE_CHILD_WINDOW, SetuArg.arg("locators", arg))
        return DefaultChildWindow(self._get_test_session(), self._get_automator(), response.get_value_for_element_setu_id())

    def LatestChildWindow(self):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.MAIN_WINDOW_GET_LATEST_CHILD_WINDOW)
        return DefaultChildWindow(self._get_test_session(), self._get_automator(), response.get_value_for_element_setu_id())

    def close_all_child_windows(self):
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.MAIN_WINDOW_CLOSE_ALL_CHILD_WINDOWS)



from .base_element import BaseElement

class GuiElement(BaseElement):
    
    def __init__(self, automator, emd, parent=None):
        super().__init__(automator, emd, parent)
        from .element_conditions import GuiElementConditions
        self.__conditions_handler = GuiElementConditions(self)

    def find(self):
        self.parent_container.find_element(self)

    identify = find

    #Override
    def find_if_not_found(self):
        if not self.is_found():
            self.find()

    def _is_partial_element(self):
        return False

    def _get_instance_number(self):
        raise Exception("Instance number is applicable only for partial gui elements.")

    def __append_instance_number(self, d):
        if self._is_partial_element():
            d["isInstanceAction"] = True
            d["instanceIndex"] = self._get_instance_number()
        return d

    def _kwargs(self, **kwargs):
        return self.__append_instance_number(kwargs)

    def _noargs(self):
        return self.__append_instance_number({})

    def _only_send_text(self, text):
        self.dispatcher.send_text(text)

    def _only_click(self):
        self.dispatcher.click()

    def __return_attr_value(self, result):
        return result and result or None

    def get_tag_name(self):
        self.find_if_not_found()
        return self.__return_attr_value(self.dispatcher.get_tag_name())

    def get_attr_value(self, attr, optional=False):
        self.find_if_not_found()
        return self.__return_attr_value(self.dispatcher.get_attr_value(attr, optional))

    def get_source(self):
        return self.get_attr_value("outerHTML")

    def get_text_content(self):
        self.find_if_not_found()
        return self.__return_attr_value(self.dispatcher.get_text_content())

    def click(self):
        self.find_if_not_found()
        self.wait_until_clickable()
        self._only_click()

    def __conditional_selected_state_click(self, condition_state):
        self.find_if_not_found()
        selected = self.is_selected()
        if selected == condition_state:
            self.wait_until_clickable()
            self._only_click()

    def select(self):
        self.__conditional_selected_state_click(False)

    def deselect(self):
        self.__conditional_selected_state_click(True)

    wait_until_present = find

    def wait_until_visible(self):
        self.find_if_not_found()
        self.conditions.IsVisible().wait()

    def wait_until_clickable(self):
        self.find_if_not_found()
        self.conditions.IsClickable().wait()

    def wait_until_selected(self):
        self.find_if_not_found()
        self.conditions.IsVisible().wait()

    #################################
    ### State Checking
    #################################
    def is_selected(self):
        self.find_if_not_found()
        return self.dispatcher.is_selected()

    def is_visible(self):
        self.find_if_not_found()
        return self.dispatcher.is_visible()

    def is_clickable(self):
        self.find_if_not_found()
        return self.dispatcher.is_clickable()

    @property
    def conditions(self):
        return self.__conditions_handler

    def _only_clear_text(self):
        self.dispatcher.clear_text()

    def _only_enter_text(self, text):
        self._only_send_text(text)

    #################################
    ### Textbox abstraction
    #################################

    def clear_text(self):
        self.find_if_not_found()
        self.wait_until_clickable()
        self._only_click()
        self._only_clear_text()

    def enter_text(self, text):
        self.find_if_not_found()
        self.wait_until_clickable()
        self._only_click()
        self._only_enter_text(text)

    def set_text(self, text):
        self.find_if_not_found()
        self.wait_until_clickable()
        self._only_click()
        self._only_clear_text()
        self._only_enter_text(text)

    def has_entered_text(self, text):
        pass

    def has_entered_partial_text(self, text):
        pass

    #################################
    ### Checkbox abstraction
    #################################
    def check(self):
        self.select()

    def uncheck(self):
        self.deselect()

    def toggle_checkbox(self):
        self.click()

    def is_checked(self):
        return self.is_selected()



import inspect

class SeleniumDriverElement:

    def __init__(self, automator_setu_id, element_setu_id, requester):
        self.__automator_setu_id = automator_setu_id
        self.__element_set_id = element_setu_id
        self.__requester = requester
        self.__partial = False
        self.__instance_index = 0

    def set_partial(self, index):
        self.__partial = True
        self.__instance_index = index

    @property
    def automator_setu_id(self):
        return self.__automator_setu_id

    @property
    def element_setu_id(self):
        return self.__element_set_id

    @property
    def requester(self):
        return self.__requester

    def post(self, **kwargs):
        input_dict = {
            "automatorSetuId" : self.automator_setu_id,
            "elementSetuId" : self.element_setu_id
        }

        if self.__partial:
            input_dict.update( {
                "isInstanceAction" : True,
                "instanceIndex" : self.__instance_index
            })
        input_dict.update(kwargs)
        json_dict = {
            "action" : inspect.stack()[1].function,
            "args" : input_dict
        }
        return self.__requester.post("/element/action", json_dict)

    def __get_value_from_response(self, response, key, optional=False):
        if "data" not in response:
            if not optional:
                raise Exception("Setu actor json response: {} does not contain 'data' key.".format(response))
            else:
                return None
        elif key not in response["data"]:
            if not optional:
                raise Exception("Setu actor json response: {} does not contain key <{}> in data section.".format(response, key))
            else:
                return None
        else:
            return response["data"][key]

    def __get_result(self, response, optional=False):
        return self.__get_value_from_response(response, "result", optional)

    def find_element(self, child_gui_element_set_id, with_type, with_value):
        self.post(
            childElementSetuId = child_gui_element_set_id,
            withType = with_type,
            withValue = with_value
        )

    def find_multielement(self, child_gui_element_set_id, with_type, with_value):
        response = self.post(
            childElementSetuId = child_gui_element_set_id,
            withType = with_type,
            withValue = with_value
        )

        return response["data"]["instanceCount"]

    def click(self):
        self.post()

    def clear_text(self):
        self.post()

    def send_text(self, text):
        self.post(text=text)

    def is_selected(self):
        response = self.post()
        return self.__get_result(response)

    def is_visible(self):
        response = self.post()
        return self.__get_result(response)

    def is_clickable(self):
        response = self.post()
        return self.__get_result(response)

    def get_tag_name(self):
        response = self.post()
        return self.__get_result(response)

    def get_attr_value(self, attr_name, optional=False):
        response = self.post(attr=attr_name)
        return self.__get_result(response, optional)

    def get_text_content(self):
        response = self.post()
        return self.__get_result(response)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .base import *

from arjuna.lib.thread.decorators import *
from arjuna.unitee.types.containers import *
from arjuna.unitee.test.defs.deps import *
from arjuna.unitee.test.defs.fixture import *
from arjuna.lib.utils import sys_utils
from arjuna.unitee.loader.deptree import DepTreeBuilder

class ModDef(TestSourceDef):
    def __init__(self, pkg, module, qname):
        super().__init__()
        self.__dep_tree_builder = DepTreeBuilder()
        self.lock = threading.RLock()
        self.pkg = pkg
        self.module = module
        self.qname = qname
        self.__threads = 1
        self.tvars = TestVars()
        self.tvars.info.module = ModuleInfo()
        self.tvars.info.module.meta['pkg'] = self.pkg
        self.tvars.info.module.meta['name'] = self.module
        self.tvars.info.module.meta['qname'] = self.qname

        self.__dependency = None
        self.__fixtures = FixturesDef()

        # For checking whether a method name exists in a given class
        self.fnames = set()
        # For pulling out method definitions by name
        self.__fdefs = {}

        self.__skipped = []
        self.__unpicked = []

        self.__fqueue = []
        # This is what would be got by groups for pickers processing. If a group picks up something, it calls setPicked()
        self.for_picker_processing = []
        # The following gets populated from above, if classDef.isNotPickedByAnyGroup() is True
        self.for_processor = []

    @property
    def threads(self):
        return self.__threads

    @threads.setter
    def threads(self, count):
        self.__threads = count

    @property
    def fixture_defs(self):
        return self.__fixtures

    def set_evars(self, variables):
        if variables:
            self.tvars.evars.update(variables.get_dict())

    @property
    def dependency(self):
        return self.__dependency

    @dependency.setter
    def dependency(self, dep):
        if not dep: return
        if not isinstance(dep, Dependency):
            raise Exception ("exclude_if attribute expects the value to be a dependency object.")
        elif not dep.is_mdep():
            raise Exception("exclude_if attribute in @tmodule expects the dependency object of modules().")
        else:
            self.__dependency = dep

    def process_dependencies(self):
        for name in self.__fqueue:
            fdef = self.__fdefs[name]
            if fdef.dependency:
                fdef.dependency.process()
                for du in fdef.dependency:
                    if du.name not in self.__fdefs:
                        du.set_ignored()
            self.__dep_tree_builder.create_node(name, fdef.dependency)

        self.__dep_tree_builder.process_dependencies()
        self.__dep_tree_builder.validate()


    def slot_func_names(self, names):
        return self.__dep_tree_builder.slot_names(names)

    def add_fixture_func(self, ftype, func):
        self.__fixtures.add_fixture_func(ftype, func)

    def get_fixture_name(self, ftype):
        return self.__fixtures.get_fixture_name(ftype)

    def get_func_names(self):
        return self.__fqueue

    def get_fdef(self, name):
        return self.__fdefs[name]

    def __iter__(self):
        return iter(self.__fdefs.items())

    def __should_include(self, fname, unpicked_set, skip_set):
        if fname in unpicked_set:
            return False
        if fname in skip_set:
            return False
        return True

    def process_unpicked_and_skipped(self):
        temp1 = set()
        for fname, fdef in self:
            if fdef.unpicked:
                self.__unpicked.append((fdef.pkg, fdef.module, fdef.name))
                temp1.add(fdef.name)

        temp2 = set()
        for fname, fdef in self:
            if fdef.skipped:
                self.__skipped.append((fdef.pkg, fdef.module, fdef.name))
                temp2.add(fdef.name)

        self.__fdefs = {i:j for i, j in self.__fdefs.items() if self.__should_include(i, temp1, temp2)}
        self.__fqueue = [i for i in self.__fqueue if i in self.__fdefs]

    def register_fdef(self, fdef):
        name = fdef.name
        self.logger.debug("Registering function definition for: {}".format(name))

        self.fnames.add(name)
        self.__fdefs[name] = fdef
        self.__fqueue.append(name)

    def enumerate(self):
        self.console.marker(50, '+')
        self.console.display(self.tvars)
        self.console.marker(50, '+')
        names = list(self.__fdefs.keys())
        names.sort()
        for name in names:
            fd = self.__fdefs[name]
            fd.enumerate()

    def should_run_func(self, qname):
        return qname in self.__fqueue


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import re
import os
import xml.etree.ElementTree as ETree
#
# from arjuna.tpi.enums import ArjunaOption
# from arjuna.lib.enums import *
# from arjuna.unitee.enums import *
from arjuna.lib.reader.hocon import *
from arjuna.lib.reader.textfile import *
from arjuna.unitee.types.root import *
from arjuna.unitee.test.defs.fixture import *

from arjuna.lib.utils import sys_utils
from arjuna.lib.utils import etree_utils
from arjuna.unitee.utils import run_conf_utils
from arjuna.unitee.selection.picker import Picker
from arjuna.unitee.selection.rules.xml_rules import XmlRules
from arjuna.lib.config import ConfigContainer
from arjuna.unitee.exceptions import *


class GroupConf(Root):

    def __init__(self, name, gconf_xml, fpath):
        super().__init__()
        self.name = name
        self.config = ConfigContainer()
        self.picker = None
        self.__rules = None
        self.threads = 1
        self.fpath = fpath
        self.root = gconf_xml
        self.__fixtures = FixturesDef()
        self.__process()

    @property
    def fixture_defs(self):
        return self.__fixtures

    @property
    def rules(self):
        return self.__rules

    def __process(self):
        def display_err_and_exit(msg):
            self.console.display_error((msg + " Fix group template file: {}").format(self.fpath))
            sys_utils.fexit()

        node_dict = etree_utils.convert_to_cidict(self.root)

        # Validate only group keys exist.
        if 'pickers' not in node_dict:
            display_err_and_exit(">>group<< element must contain >>pickers<< element.")

        for child_tag, child in node_dict.items():
            child_tag = child_tag.lower()
            if child_tag == 'config':
                config = child
                for option in config:
                    run_conf_utils.validate_config_xml_child("groups", self.fpath, option)
                    run_conf_utils.add_config_node_to_configuration("groups", self.config, option)
            elif child_tag == 'fixtures':
                fixtures = child
                for child in fixtures:
                    run_conf_utils.validate_fixture_xml_child("groups", "group", self.fpath, child)
                    run_conf_utils.add_fixture_node_to_fixdefs(self.fixture_defs, child)
            elif child_tag =='pickers':
                self.picker = Picker(self.fpath, self.name, child)
            elif child_tag =='rules':
                self.__rules = XmlRules(self.fpath, self.name, child)
            else:
                display_err_and_exit("Unexpected element >>{}<< found in session definition.".format(child.tag))


class GroupConfsLoader:

    @staticmethod
    def __load_pick_all(gconfs):
        from arjuna.tpi import Arjuna
        central_config = Arjuna.get_central_config()
        arjuna_root_dir = central_config.get_arjuna_option_value(ArjunaOption.ARJUNA_ROOT_DIR).as_string()
        fpath = os.path.join(
            arjuna_root_dir,
            "res/st/magroup.xml"
        )
        group_xml = ETree.parse(fpath).getroot()
        group_name = group_xml.attrib['name']
        gconfs[group_name] = GroupConf(group_name, group_xml, fpath)

    @staticmethod
    def __load_user_gconfs(gconfs):
        from arjuna.tpi import Arjuna
        console = Arjuna.get_console()
        central_config = Arjuna.get_central_config()
        ugcdir = central_config.get_arjuna_option_value(ArjunaOption.CONFIG_DIR).as_string()
        ugfpath = os.path.join(ugcdir, "groups.xml")

        def display_err_and_exit(msg):
            console.display_error((msg + " Fix groups template file: {}").format(ugfpath))
            sys_utils.fexit()

        if not os.path.exists(ugfpath) or not os.path.isfile(ugfpath):
            return

        try:
            tree = ETree.parse(ugfpath)
        except Exception as e:
            print(e)
            display_err_and_exit("Groups definition file could not be loaded because of errors in XML.")
        else:
            root = tree.getroot()
            if root.tag != 'groups':
                display_err_and_exit("Invalid groups template file. Root element tag should be >>groups<<.")
            node_dict = etree_utils.convert_to_cidict(root)

            # Validate only group keys exist.
            if node_dict.keys() != {'group'}:
                display_err_and_exit(">>groups<< element can contain only one or more >>group<< elements.")
            elif not node_dict:
                display_err_and_exit(">>groups<< element must contain atleast one >>group<< element.")
            else:
                for group in list(root):
                    run_conf_utils.validate_group_xml_child("session", ugfpath, group)
                    group_attrs = etree_utils.convert_attribs_to_cidict(group)
                    name = group_attrs['name'].strip()
                    if not name:
                        display_err_and_exit(">>name<< attribute in group definition can not be empty.")
                    gconfs[name] = GroupConf(name, group, ugfpath)

    @staticmethod
    def load():
        from arjuna.unitee.types.containers import CIStringDict
        gconfs = CIStringDict()
        GroupConfsLoader.__load_pick_all(gconfs)
        GroupConfsLoader.__load_user_gconfs(gconfs)
        return gconfs



